# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Generator, Iterator, List, Optional, Sequence, Tuple, Union

import torch

from megatron.energon.edataclass import edataclass
from megatron.energon.flavors.base_dataset import FlexState, SavableDataset
from megatron.energon.flavors.webdataset.itar_reader import ITarReader
from megatron.energon.flavors.webdataset.structs import FilteredSample
from megatron.energon.rng import WorkerRng
from megatron.energon.worker import WorkerConfig


@edataclass
class RawSampleData:
    """Represents the iteration state of a single slice slice to the index."""

    #: Index of the sample. This is also the restore key
    __restore_key__: Tuple[str, int]
    #: The sample data
    data: Tuple[Optional[FilteredSample], ...]


@edataclass
class SliceState:
    """Represents the iteration state of a single slice slice to the index."""

    #: The slice index of this slice state
    index: int
    #: The actual state: The global sample offset (`slice[index] <= offset < slice[index + 1]``)
    current: int


class WebdatasetSampleLoaderDataset(SavableDataset[RawSampleData]):
    """Internal class for loading samples from webdataset slices"""

    #: The readers for each joined dataset
    join_readers: Sequence[ITarReader]

    #: The offsets of the slice slices to iterate over for the current worker
    slice_offsets: Optional[Sequence[int]]

    # If = 1, every sample is seen exactly once per epoch. If > 1, samples
    # (or rather slice slices) are shuffled within this number of epochs (i.e. randomly
    # selected without replacement). If None, the slices are effectively shuffle over
    # infinite epochs (i.e. slice slices are drawn with replacement).
    shuffle_over_epochs: Optional[int]
    # Number of parallel iterators to be opened simultaneously (and random sample between them)
    parallel_slice_iters: int

    # Worker's random generator
    _worker_rng: WorkerRng

    #: The RNG state to be used for regenerating the pending slices
    _pending_slices_rng_state: Optional[FlexState]
    #: The number of slices that have already been opened / processed and thus been removed from the
    # pending slices.
    _pending_slices_offset: Optional[int]
    #: Pending slices are the slices which have not yet been opened, but should be processed
    # in the current "epoch". If None, regenerate from the seed and offset.
    _pending_slice_indexes: Optional[List[int]]
    #: The active slices are the currently opened slices. May contain `None`, if there are fewer
    # slices available (i.e. pending_slices empty) than parallel slice iterators requested.
    _active_slice_state: List[Optional[SliceState]]
    #: The total number of samples retrieved, it's just a monotonically increasing counter
    _sample_count: int
    #: Number of epochs this dataset has been iterated over
    _epoch_count: int
    #: The number of samples retrieved in current epoch
    _epoch_sample_count: int

    _savable_fields = (
        "_worker_rng",
        "_pending_slices_offset",
        "_pending_slice_indexes",
        "_active_slice_state",
        "_sample_count",
        "_epoch_count",
        "_epoch_sample_count",
    )

    def __init__(
        self,
        join_readers: Sequence[ITarReader],
        workers_sample_slice_offsets: Sequence[Sequence[int]],
        *,
        worker_config: WorkerConfig,
        shuffle_over_epochs: Optional[int] = None,
        parallel_slice_iters: int = 1,
    ):
        """
        The webdataset loader. Iterates over the slice infos and yields the samples.

        Args:
            join_readers: A sequence of the joined readers (or just a single reader) to iterate over.
            worker_slice_offsets: The offsets of the slice slices to iterate over, for each worker.
            worker_config: The worker configuration.
            shuffle_over_epochs: If None, disable shuffling.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather slice slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the slices are effectively shuffle over infinite epochs (i.e. slice slices
                are drawn with replacement).
            parallel_slice_iters: If > 1, samples are randomly drawn from parallel slice iterators.
                This will not impact performance, but increase randomness. If = 1, the slices are
                iterated in order.
        """
        super().__init__(worker_config=worker_config)

        self.join_readers = join_readers
        self.shuffle_over_epochs = shuffle_over_epochs
        self.parallel_slice_iters = parallel_slice_iters

        # Store the slices for all workers
        # The slices for the current worker, will have to be extracted from this list later
        self.workers_slice_offsets = workers_sample_slice_offsets
        self.slice_offsets = None

        self.reset_state_own()

        assert shuffle_over_epochs is None or shuffle_over_epochs == -1 or shuffle_over_epochs >= 1
        assert self.parallel_slice_iters >= 1

    def reset_state_own(self) -> None:
        self._worker_rng = WorkerRng(self.worker_config)
        self._pending_slice_indexes = None
        self._pending_slices_offset = None
        self._pending_slices_rng_state = None
        self._active_slice_state = [None] * self.parallel_slice_iters
        self._sample_count = 0
        self._epoch_count = 0
        self._epoch_sample_count = 0

    def ensure_slice_offsets(self) -> None:
        self.worker_config.assert_worker()

        if self.slice_offsets is None:
            self.slice_offsets = self.workers_slice_offsets[self.worker_config.rank_worker_id()]

    def _get_sample(self, index: int) -> RawSampleData:
        return RawSampleData(
            __restore_key__=("Webdataset", index),
            data=tuple(reader[index] for reader in self.join_readers),
        )

    def _slices_once(self) -> List[int]:
        """Yields the indexes to slice offsets once. Possibly shuffles the list."""
        assert self.slice_offsets is not None

        num_slices = len(self.slice_offsets) - 1
        slices_offset = self._pending_slices_offset

        if self.shuffle_over_epochs is None:
            # No shuffling
            res_list = list(range(num_slices))
            if slices_offset is None:
                slices_offset = 0
        else:
            # Restore state or start new (and save)
            if slices_offset is None:
                # Start new state. First, save the state to restore the same order.
                self._pending_slices_rng_state = self._worker_rng.save_state()
                rng = self._worker_rng
                slices_offset = 0
            else:
                # Restore the state. Create a dedicated rng for this, as the main rng is in the
                # state for iterating from the next iterator.
                assert self._pending_slices_rng_state is not None
                rng = WorkerRng(self.worker_config)
                rng.restore_state(self._pending_slices_rng_state)

            if self.shuffle_over_epochs == -1:
                # Shuffle with replacement (i.e. infinite epochs), effectively return as many slices
                # as are required for parallel slice iterators.
                # Next slices are drawn in the _slices_iter.
                res_list = [rng.randbelow(num_slices) for _ in range(self.parallel_slice_iters)]
            elif self.shuffle_over_epochs >= 1:
                # Shuffle without replacement (potentially over multiple epochs)
                res_list = rng.shuffle(list(range(num_slices)) * self.shuffle_over_epochs)
            else:
                raise ValueError(f"Invalid shuffle_over_epochs: {self.shuffle_over_epochs}")
        # Reverse, such that pop returns the first element (in O(1) time)
        res_list.reverse()
        # Skip restored slice list already processed slices
        assert slices_offset is not None
        self._pending_slices_offset = slices_offset
        if slices_offset > 0:
            # Those have already been popped in the current state
            del res_list[-slices_offset:]
        # Set the pending slices
        self._pending_slice_indexes = res_list
        return res_list

    def _slices_iter(self) -> Generator[RawSampleData, None, None]:
        """Iterates the samples in a list of slices, possibly using multiple parallel iterators over
        the slices."""

        assert self.slice_offsets is not None

        active_slice_probs = torch.zeros(self.parallel_slice_iters, dtype=torch.float32)
        active_slices = self._active_slice_state
        pending_slice_indexes = self._pending_slice_indexes

        def slice_at(idx: int) -> SliceState:
            assert self.slice_offsets is not None
            return SliceState(
                index=idx,
                current=self.slice_offsets[idx],
            )

        # Weight the slices by their size to get a more even distribution of samples
        if any(s is not None for s in active_slices) or self._pending_slices_offset is not None:
            # Having an active state, or pending slices. This means we are resuming an epoch.
            if pending_slice_indexes is None:
                # Need to restore the pending slices
                pending_slice_indexes = self._slices_once()
            assert pending_slice_indexes is not None

            # Restore the state
            assert len(active_slices) == self.parallel_slice_iters
            for idx, slice_state in enumerate(active_slices):
                if slice_state is not None:
                    active_slice_probs[idx] = (
                        self.slice_offsets[slice_state.index + 1]
                        - self.slice_offsets[slice_state.index]
                    )

            if self.worker_config.should_log(level=1):
                self.worker_config.worker_log(
                    {
                        "t": "WebdatasetSampleLoaderDataset._slices_iter.resume_epoch",
                        "r": self.worker_config.rank,
                        "w": self.worker_config.rank_worker_id(),
                        "pending_slice_indexes": pending_slice_indexes,
                        "active_slices": [
                            (
                                None
                                if state is None
                                else {
                                    "index": state.index,
                                    "current": state.current,
                                }
                            )
                            for state in active_slices
                        ],
                        "count": self._sample_count,
                        "epoch": self._epoch_count,
                        "epoch_count": self._epoch_sample_count,
                        "probs": active_slice_probs.tolist(),
                    }
                )

        else:
            # Start a new epoch
            assert pending_slice_indexes is None
            pending_slice_indexes = self._slices_once()

            if self.worker_config.should_log(level=1):
                self.worker_config.worker_log(
                    {
                        "t": "WebdatasetSampleLoaderDataset._slices_iter.next_epoch",
                        "r": self.worker_config.rank,
                        "w": self.worker_config.rank_worker_id(),
                        "pending_slice_indexes": pending_slice_indexes,
                        "count": self._sample_count,
                        "epoch": self._epoch_count,
                        "epoch_count": self._epoch_sample_count,
                        "probs": active_slice_probs.tolist(),
                        "shuffle_over_epochs": self.shuffle_over_epochs,
                    }
                )

            assert self._pending_slices_offset is not None

            # List of slice iterators, always of length `parallel_slice_iters`. May contain `None`.
            active_slices.clear()
            # Fill up the slice iterators
            while len(pending_slice_indexes) > 0 and len(active_slices) < self.parallel_slice_iters:
                slice_index = pending_slice_indexes.pop()
                self._pending_slices_offset += 1
                slice_state = slice_at(slice_index)
                active_slice_probs[len(active_slices)] = (
                    self.slice_offsets[slice_state.index + 1]
                    - self.slice_offsets[slice_state.index]
                )
                active_slices.append(slice_state)
            # Fill up the slice iterators with None
            for _ in range(len(active_slices), self.parallel_slice_iters):
                active_slices.append(None)

        # print(
        #     f"Next slice iters generated for {self.worker_config.rank}:{self.worker_config.rank_worker_id()}: probs={active_slice_probs}"
        # )
        # for slice_state in active_slices:
        #     if slice_state is None:
        #         print("  - None")
        #     else:
        #         print(
        #             f"  - [{slice_offsets[slice_state.index]}, {slice_offsets[slice_state.index + 1]}] at {slice_state.current}"
        #         )

        # Iterate over the slice iterators while there is an iterator left
        while torch.count_nonzero(active_slice_probs).item() > 0:
            if self.shuffle_over_epochs is None:
                # No shuffling, deterministic order, always the same
                assert self.parallel_slice_iters == 1
                slice_idx = 0
            else:
                # Take a random slice iterator
                slice_idx = self._worker_rng.choice_idx(active_slice_probs)
            slice_state = active_slices[slice_idx]
            assert slice_state is not None
            sample = self._get_sample(slice_state.current)
            # print(f"Read sample at {slice_state.current} -> {'None' if sample is None or sample.data[0] is None else sample.data[0]['__key__']}")
            slice_state.current += 1
            self._sample_count += 1
            self._epoch_sample_count += 1
            if slice_state.current >= self.slice_offsets[slice_state.index + 1]:
                # Iterator exhausted -> take next / remove from list
                if len(pending_slice_indexes) > 0 or self.shuffle_over_epochs == -1:
                    if len(pending_slice_indexes) > 0:
                        # Take the next slice (without replacement)
                        next_idx = pending_slice_indexes.pop()
                        assert self._pending_slices_offset is not None
                        self._pending_slices_offset += 1
                    else:
                        # Randomly select a new slice directly (with replacement)
                        num_slices = len(self.slice_offsets) - 1
                        next_idx = self._worker_rng.randbelow(num_slices)
                    next_slice_state = slice_at(next_idx)
                    active_slice_probs[slice_idx] = (
                        self.slice_offsets[next_slice_state.index + 1]
                        - self.slice_offsets[next_slice_state.index]
                    )
                    active_slices[slice_idx] = next_slice_state
                    # print(
                    #     f"Slice iter for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} "
                    #     f"[{slice_offsets[slice_state.index]}, {slice_offsets[slice_state.index + 1]}] exhausted at {slice_state.current}, "
                    #     f"taking next slice {next_slice_state} [{slice_offsets[next_slice_state.index]}, {slice_offsets[next_slice_state.index + 1]}], "
                    #     f"{len(pending_slice_indexes)} slices left, probs={active_slice_probs.tolist()}"
                    # )
                else:
                    active_slice_probs[slice_idx] = 0
                    active_slices[slice_idx] = None
                    # print(
                    #     f"Slice iter for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} "
                    #     f"[{slice_offsets[slice_state.index]}, {slice_offsets[slice_state.index + 1]}] exhausted at {slice_state.current}, "
                    #     f"no next slice, probs={active_slice_probs.tolist()}"
                    # )
                if self.worker_config.should_log(level=2):
                    self.worker_config.worker_log(
                        {
                            "t": "WebdatasetSampleLoaderDataset._slices_iter.exhausted",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "remaining": len(pending_slice_indexes),
                            "count": self._sample_count,
                            "epoch": self._epoch_count,
                            "epoch_count": self._epoch_sample_count,
                            "probs": active_slice_probs.tolist(),
                        }
                    )
            if sample.data[0] is not None:
                # Otherwise the sample was skipped.
                if self.worker_config.should_log(level=1):
                    self.worker_config.worker_log(
                        {
                            "t": "WebdatasetSampleLoaderDataset._slices_iter.yield",
                            "r": self.worker_config.rank,
                            "w": self.worker_config.rank_worker_id(),
                            "index": sample.__restore_key__[1],
                            "key": sample.data[0]["__key__"],
                            "shard": sample.data[0]["__shard__"],
                            "count": self._sample_count,
                            "epoch": self._epoch_count,
                            "epoch_count": self._epoch_sample_count,
                        }
                    )
                # Now, yield the sample
                yield sample
                del sample
        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset._slices_iter.all_exhausted",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "count": self._sample_count,
                    "epoch": self._epoch_count,
                    "epoch_count": self._epoch_sample_count,
                }
            )

        # Epoch has finished, reset states.
        self._epoch_count += 1
        self._epoch_sample_count = 0
        self._pending_slice_indexes = None
        self._pending_slices_offset = None
        # print(
        #     f"slice iters exhausted for {self.worker_config.rank}:{self.worker_config.rank_worker_id()} after {cnt} samples"
        # )

    def __len__(self) -> int:
        return sum(
            slice_offsets[-1] - slice_offsets[0] for slice_offsets in self.workers_slice_offsets
        )

    def worker_has_samples(self) -> bool:
        self.worker_config.assert_worker()
        self.ensure_slice_offsets()
        assert self.slice_offsets is not None
        return len(self.slice_offsets) > 1

    def __iter__(self) -> Iterator[RawSampleData]:
        self.worker_config.assert_worker()

        self.ensure_slice_offsets()
        assert self.slice_offsets is not None

        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "WebdatasetSampleLoaderDataset.__iter__",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "slice_offsets": self.slice_offsets,
                    "parallel_slice_iters": self.parallel_slice_iters,
                    "shuffle_over_epochs": self.shuffle_over_epochs,
                }
            )

        if len(self.slice_offsets) <= 1:
            return

        yield from self._slices_iter()

    def can_restore_sample(self) -> bool:
        return True

    def assert_can_restore(self) -> None:
        pass

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> RawSampleData:
        # Key is: ("Webdataset", index)
        # The key is joined in the dataset's typed joining (i.e. load_sample of JoinedWebdatasetFactory).
        id, index = restore_key
        assert id == "Webdataset"
        assert isinstance(index, int)
        return self._get_sample(index)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "joins": len(self.join_readers),
            "len": len(self),
            "slice_offsets": [
                list(worker_slice_offsets) for worker_slice_offsets in self.workers_slice_offsets
            ],
            "worker_config": self.worker_config.config(),
            "shuffle_over_epochs": self.shuffle_over_epochs,
            "parallel_slice_iters": self.parallel_slice_iters,
        }

    def __str__(self):
        return f"WebdatasetSampleLoaderDataset(join_readers={self.join_readers}, shuffle_over_epochs={self.shuffle_over_epochs}, parallel_slice_iters={self.parallel_slice_iters})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

"""Main entry point for energon CLI tools."""

import click

from megatron.energon.tools.analyze_debug import command as analyze_debug_command
from megatron.energon.tools.checkpoint import command as checkpoint_command
from megatron.energon.tools.info import command as info_command
from megatron.energon.tools.lint import command as lint_command
from megatron.energon.tools.mount import command as mount_command
from megatron.energon.tools.prepare import command as prepare_command
from megatron.energon.tools.preview import command as preview_command


def print_version(ctx, param, value):
    """Print version and exit."""
    if not value or ctx.resilient_parsing:
        return
    try:
        from megatron.energon import __version__

        click.echo(f"megatron-energon {__version__}")
    except ImportError:
        click.echo("megatron-energon version unknown")
    ctx.exit()


@click.group(context_settings=dict(help_option_names=["-h", "--help"]), invoke_without_command=True)
@click.option(
    "--version",
    "-V",
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=print_version,
    help="Show version and exit.",
)
@click.pass_context
def main(ctx):
    """A set of tools that energon provides.

    Among other things, you can use it to lint or preprocess your dataset.

    See help of commands to learn more."""

    # This is needed to show help if no subcommand is provided
    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))


main.add_command(analyze_debug_command)
main.add_command(checkpoint_command)
main.add_command(lint_command)
main.add_command(info_command)
main.add_command(mount_command)
main.add_command(prepare_command)
main.add_command(preview_command)

if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import sqlite3
import struct
from pathlib import Path
from typing import Any, BinaryIO, Generator, List, Optional, Tuple, Union

from numpy import int8

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.webdataset.itar import ITarSamplePointer
from megatron.energon.flavors.webdataset.thread_local_sqlite import ThreadLocalSqlite
from megatron.energon.local_copy import ensure_local_copy


class SqliteIndexWriter:
    sqlite_path: EPath
    db: Optional[sqlite3.Connection]
    duplicates: List[Tuple[str, int]]

    def __init__(self, sqlite_path: EPath):
        """
        Initializes an SQLite database and sets up the samples table:
          - samples(tar_file_id INTEGER,
                    sample_key TEXT,
                    sample_index INTEGER,
                    byte_offset INTEGER,
                    byte_size INTEGER)
        Also creates an index on samples(sample_key).
        """

        # Final path and temporary path
        self.sqlite_path = sqlite_path

        # Initialize SQLite connection
        path = str(self.sqlite_path)
        # Only supporting local file system, because sqlite does not support remote file systems.
        # TODO: Implement remote file systems. Maybe create locally in tmp then upload?
        assert path.startswith("/"), (
            f"SQLite path must be absolute local file system path: {self.sqlite_path}"
        )
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        self.db = sqlite3.connect(path)
        self.db.execute("PRAGMA busy_timeout = 5000;")  # wait up to 5000ms when locked
        self.db.execute("PRAGMA journal_mode = WAL;")

        # Create the sample table
        self.db.execute("DROP INDEX IF EXISTS idx_samples_sample_key")
        self.db.execute("DROP INDEX IF EXISTS idx_samples_by_tar_and_idx")
        self.db.execute("DROP TABLE IF EXISTS samples")
        self.db.execute(
            """
            CREATE TABLE samples (
                tar_file_id INTEGER,
                sample_key TEXT,
                sample_index INTEGER,
                byte_offset INTEGER,
                byte_size INTEGER
            )
        """
        )

        # Create the sample parts table
        self.db.execute("DROP INDEX IF EXISTS idx_sample_parts_seq")
        self.db.execute("DROP INDEX IF EXISTS idx_sample_parts_full")
        self.db.execute("DROP TABLE IF EXISTS sample_parts")
        self.db.execute(
            """
            CREATE TABLE sample_parts (
                tar_file_id INTEGER,
                sample_index INTEGER,
                part_name TEXT,
                content_byte_offset INTEGER,
                content_byte_size INTEGER
            )
        """
        )

        self.duplicates = []

    def append_sample(
        self,
        tar_file_id: int8,
        sample_key: str,
        sample_index: int,
        byte_offset: Optional[int],
        byte_size: Optional[int],
    ):
        """
        Adds a new sample row to the samples table.

        Args:
            tar_file_id: The index of the tar file in the reader.
            sample_key: The key of the sample.
            sample_index: The index of the sample in the tar file.
            byte_offset: The byte offset of the sample in the tar file.
            byte_size: The size of the sample in the tar file.
        """

        assert self.db is not None, "Database is closed"

        # Insert a row in the samples table
        self.db.execute(
            """
            INSERT INTO samples (tar_file_id, sample_key, sample_index, byte_offset, byte_size)
            VALUES (?, ?, ?, ?, ?)
            """,
            (tar_file_id, sample_key, sample_index, byte_offset, byte_size),
        )

    def append_part(
        self,
        tar_file_id: int8,
        sample_index: int,
        part_name: str,
        content_byte_offset: int,
        content_byte_size: int,
    ):
        """Adds a new part row to the samples table."""

        assert self.db is not None, "Database is closed"

        # Insert a row in the sample parts table
        self.db.execute(
            """
            INSERT INTO sample_parts (tar_file_id, sample_index, part_name, content_byte_offset, content_byte_size)
            VALUES (?, ?, ?, ?, ?)
            """,
            (tar_file_id, sample_index, part_name, content_byte_offset, content_byte_size),
        )

    def close(self):
        """
        Closes the DB connection. If finalize=True, the temporary database is
        renamed to the final name, overwriting if necessary.
        """
        assert self.db is not None, "Database is closed"

        # Create the index after adding all the samples for better speed
        # Index on sample_key for fast lookups
        self.db.execute("CREATE INDEX IF NOT EXISTS idx_samples_sample_key ON samples(sample_key)")

        # Create index on the samples table.  Help the planner if it chooses `samples` as the probe side of the join
        self.db.execute(
            "CREATE INDEX IF NOT EXISTS idx_samples_by_tar_and_idx ON samples(tar_file_id, sample_index)"
        )

        # Create index on the sample_parts table for fast sequential access
        self.db.execute(
            "CREATE INDEX IF NOT EXISTS idx_sample_parts_seq ON sample_parts(tar_file_id, sample_index, content_byte_offset)"
        )

        # Create a full index on the sample_parts table for equality lookups and getting offsets directly from key
        self.db.execute(
            "CREATE INDEX IF NOT EXISTS idx_sample_parts_full ON sample_parts(tar_file_id, sample_index, part_name, content_byte_offset, content_byte_size)"
        )

        # Check if sample_key are all unique
        # self.db.execute("CREATE TEMP TABLE temp AS SELECT sample_key, COUNT(*) AS c FROM samples GROUP BY sample_key HAVING c > 1")
        duplicates = self.db.execute(
            "SELECT sample_key, COUNT(*) AS c FROM samples GROUP BY sample_key HAVING c > 1 LIMIT 5"
        ).fetchall()
        if len(duplicates) > 0:
            self.duplicates = duplicates

        if self.db is not None:
            self.db.commit()
            self.db.close()
            self.db = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # If an exception occurred, do not finalize (so you can inspect the temp file)
        self.close()


class JoinIndexWriter:
    """Describes how one primary dataset is joined with multiple secondary datasets.

    For fast random access, this is a binary format that is memory-mapped.
    The first 16 bytes are a header with the number of columns (1 primary + N secondary).
    Each row contains (shard_idx, byte_offset, byte_size) for each column.
    """

    def __init__(self, join_index_path: EPath):
        self.join_index_path = join_index_path
        self.join_index_file = join_index_path.open("wb")
        self.num_columns = None

    def append(self, *columns: Tuple[int, int, int]):
        """Appends a new row to the join index file.

        Each row contains (shard_idx, byte_offset, byte_size) for each column.
        """

        if self.num_columns is None:
            # Write the number of columns
            self.join_index_file.write(b"JIDX0001")  # Magic bytes with version
            self.join_index_file.write(struct.pack("q", len(columns)))
            self.num_columns = len(columns)
        else:
            assert len(columns) == self.num_columns, (
                f"Inconsistent number of keys: Had {self.num_columns} before, got {len(columns)}"
            )

        # Write the columns
        for key in columns:
            assert isinstance(key, tuple) and len(key) == 3
            self.join_index_file.write(struct.pack("qqq", *key))

    def close(self):
        self.join_index_file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class SqliteIndexReader:
    """Reads samples from an SQLite database created by SqliteIndexWriter.

    The database contains a table with the following schema:
    - samples(tar_file_id INTEGER,
              sample_key TEXT,
              sample_index INTEGER,
              byte_offset INTEGER,
              byte_size INTEGER)
    """

    sqlite_path: EPath
    db: ThreadLocalSqlite

    def __init__(self, sqlite_path: EPath):
        """Initialize the SQLite database reader.

        Args:
            sqlite_path: Path to the SQLite database file
        """
        self.sqlite_path = ensure_local_copy(sqlite_path)

        # Initialize SQLite connection
        path = str(self.sqlite_path)
        # Only supporting local file system, because sqlite does not support remote file systems
        assert path.startswith("/"), (
            f"SQLite path must be absolute local file system path: {self.sqlite_path}"
        )

        self.db = ThreadLocalSqlite(path)

    def list_all_samples(self) -> Generator[Tuple[str, int, int], None, None]:
        """List all sample keys in the database.

        Returns:
            Tuple of (sample_key, byte_size)
        """

        assert self.db is not None, "Database is closed"

        for row in self.db.select_all("SELECT sample_key, byte_size, tar_file_id FROM samples"):
            yield row[0], row[1], row[2]

    def list_all_sample_parts(self) -> Generator[Tuple[str, int, int], None, None]:
        """List all sample parts (i.e. individual files) in the database.

        Returns:
            Tuple of (full_key, size, tar_file_id)
        """

        assert self.db is not None, "Database is closed"

        # Select all parts (sorted by tar_file_id, sample_index) but joined with the sample_key names
        for row in self.db.select_all(
            "SELECT "
            "s.sample_key || '.' || sp.part_name AS full_key, "
            "sp.content_byte_size AS size, "
            "sp.tar_file_id AS tar_file_id "
            "FROM sample_parts AS sp "
            "JOIN samples AS s "
            "ON sp.tar_file_id  = s.tar_file_id AND sp.sample_index = s.sample_index "
            "ORDER BY sp.tar_file_id, sp.sample_index, sp.content_byte_offset"
        ):
            yield row[0], row[1], row[2]

    def list_sample_parts(self, sample_key: str) -> Generator[Tuple[str, int, int], None, None]:
        """List all sample parts (i.e. individual files) in the database.

        Args:
            sample_key: The sample key to look up

        Returns:
            Tuple of (part_name, size, tar_file_id)
        """

        assert self.db is not None, "Database is closed"

        # Select all parts (sorted by tar_file_id, sample_index) but joined with the sample_key names
        for row in self.db.select_all(
            "SELECT "
            "sp.part_name AS part_name, "
            "sp.content_byte_size AS size, "
            "sp.tar_file_id AS tar_file_id "
            "FROM sample_parts AS sp "
            "JOIN samples AS s "
            "ON sp.tar_file_id  = s.tar_file_id AND sp.sample_index = s.sample_index "
            "WHERE s.sample_key = ? "
            "ORDER BY sp.tar_file_id, sp.sample_index, sp.content_byte_offset",
            (sample_key,),
        ):
            yield row[0], row[1], row[2]

    def get_total_size(self) -> int:
        """Get the total size of all samples in the database."""
        assert self.db is not None, "Database is closed"

        count = self.db.select_one("SELECT SUM(byte_size) FROM samples")
        return count[0] if count else 0

    def get_sample_count(self) -> int:
        """Get the total number of samples in the database."""
        assert self.db is not None, "Database is closed"

        count = self.db.select_one("SELECT COUNT(*) FROM samples")
        return count[0] if count else 0

    def get_sample_pointer_by_key(self, key: str) -> Any:
        """Get a sample by its key name.

        Args:
            key: The sample key to look up

        Returns:
            Tuple of (tar_file_id, sample_key, sample_index, byte_offset, byte_size)
        """
        assert self.db is not None, "Database is closed"

        sample = self.db.select_one(
            "SELECT tar_file_id, sample_key, sample_index, byte_offset, byte_size "
            "FROM samples WHERE sample_key = ?",
            (key,),
        )

        if sample is None:
            raise KeyError(f"Sample key not found: {key}")

        return ITarSamplePointer(
            tar_file_id=sample[0],
            byte_offset=sample[3],
            byte_size=sample[4],
        )

    def close(self):
        """Close the database connection."""
        if self.db is not None:
            self.db.thread_close()
            del self.db

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class JoinIndexReader:
    """Reads a join index file in different ways.

    If a column is specified, only that column is read, otherwise the full rows.
    You can iterate over the rows, or read a specific row by index, or get the full tensor.

    Each row contains (shard_idx, byte_offset, byte_size) for each column.
    """

    join_index_path: EPath
    join_index_file: BinaryIO
    column: Optional[int]
    num_columns: int
    has_iterated: bool
    index_row_position: int

    def __init__(self, join_index_path: EPath, column: Optional[int] = None):
        self.join_index_path = join_index_path
        self.join_index_byte_size = join_index_path.size()

        self.column = column

        self.join_index_file = join_index_path.open("rb")
        self.has_iterated = False
        self.index_row_position = -1

        # Read the header
        bytes_magic = self.join_index_file.read(8)
        assert isinstance(bytes_magic, bytes)
        assert bytes_magic[:4] == b"JIDX", f"Invalid magic bytes: {bytes_magic}"
        assert bytes_magic[4:8] == b"0001", f"Unsupported version: {bytes_magic[4:8]}"

        # Read the number of columns
        bytes_seckeys = self.join_index_file.read(8)
        assert isinstance(bytes_seckeys, bytes)
        self.num_columns = struct.unpack("q", bytes_seckeys)[0]

        self.index_row_position = 0

    def get_as_tensor(self):
        """Returns the join index as a tensor with shape (N, num_columns, 3)."""

        assert not self.has_iterated, "Cannot get_as_tensor after iterating"

        import torch

        # Read the raw bytes for all N * 3 int64s.
        data = self.join_index_file.read()
        self.index_file_position = self.join_index_file.tell()
        assert len(data) % (8 * 3) == 0, (
            f"Index file reading: Expected multiple of 3 * 8 bytes, got {len(data)} bytes"
        )

        return torch.frombuffer(data, dtype=torch.int64).view(-1, self.num_columns, 3)

    def __len__(self):
        return (self.join_index_byte_size - 16) // (self.num_columns * 8 * 3)

    def __iter__(self):
        return self

    def _read_one_row(
        self, column: Optional[int] = None
    ) -> Union[None, List[Tuple[int, int, int]]]:
        row = []
        for col_idx in range(self.num_columns):
            if column is not None and col_idx != column:
                # Skip this column
                self.join_index_file.seek(8 * 3, 1)
                continue

            bytes_key = self.join_index_file.read(8 * 3)
            if not bytes_key:
                return None
            assert isinstance(bytes_key, bytes)
            key_tuple = struct.unpack("qqq", bytes_key)
            row.append(key_tuple)
        self.index_row_position += 1
        return row

    def __next__(self) -> Union[None, List[Tuple[int, int, int]]]:
        self.has_iterated = True
        return self._read_one_row(column=self.column)

    def tell_row(self) -> int:
        return self.index_row_position

    def __getitem__(self, idx: int) -> List[Tuple[int, int, int]]:
        """Reads the idx-th row of the join index file."""

        assert 0 <= idx < len(self), f"Index out of bounds: {idx} not in [0, {len(self)})"

        # Seek to the correct position

        if self.index_row_position != idx:
            self.join_index_file.seek(16 + idx * self.num_columns * 8 * 3)
            self.index_row_position = idx

        # Read the secondary keys
        row = self._read_one_row(column=self.column)
        assert row is not None, f"Failed to read row {idx}"

        return row

    def close(self):
        self.join_index_file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import json
from typing import Any, BinaryIO, Union

import ryml

from megatron.energon.epathlib.epath import EPath


def load_yaml_json(path: EPath) -> Any:
    """Load a YAML or JSON file from a path. Loading json is much faster than yaml.

    Args:
        path: The path to the file to load.

    Returns:
        The parsed YAML or JSON data as a native Python object.
    """
    if path.name.endswith(".yaml"):
        with path.open("rb") as f:
            return load_yaml(f)
    elif path.name.endswith(".json"):
        with path.open("rb") as f:
            return json.load(f)
    else:
        raise ValueError(f"Invalid file extension: {path.name}")


def load_yaml(stream: Union[BinaryIO, bytes]) -> Any:
    """Load a YAML file from a stream or bytes object using rapidyaml/ryml.
    This is much faster than the standard yaml library, but we don't
    support all YAML features.

    Args:
        stream: A stream or bytes object containing the YAML data.

    Returns:
        The parsed YAML data as a native Python object.
    """

    if isinstance(stream, bytes):
        bytes_data = stream
    else:
        bytes_data = stream.read()
    tree = ryml.parse_in_arena(bytes_data)
    native_obj = _ryml_tree_to_native(tree)

    return native_obj


def _cast_scalar(tree: ryml.Tree, nid: int) -> object:
    """Convert YAML scalar to bool | int | float | None | str."""
    memview = tree.val(nid)
    s = memview.tobytes().decode()

    if tree.is_val_quoted(nid):
        return s

    s_lower = s.lower()

    if s_lower in {"null", "~"}:
        return None
    if s_lower in {"true", "yes"}:
        return True
    if s_lower in {"false", "no"}:
        return False
    try:
        if "." in s or "e" in s_lower:
            return float(s)
        return int(s)
    except ValueError:
        return s


def _to_native(t: ryml.Tree, nid: int):
    if t.is_map(nid):
        # iterate children of a mapping node
        return {t.key(ch).tobytes().decode(): _to_native(t, ch) for ch in ryml.children(t, nid)}
    if t.is_seq(nid):
        return [_to_native(t, ch) for ch in ryml.children(t, nid)]
    # scalar leaf
    return _cast_scalar(t, nid)


def _ryml_tree_to_native(tree: ryml.Tree):
    return _to_native(tree, tree.root_id())



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Generic, Iterator, Optional, TypeVar

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


class EpochizeDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """
    Uses the base dataset, and creates one epoch, which has length samples. Keeps the underlying
    dataset iterator alive over epochs (i.e. if it is an infinite dataset, it will keep the state).
    Repeats the underlying dataset if the iterator is exhausted.
    """

    length: int
    _active_iter: Optional[Iterator[T_sample]]
    _offset: int

    _savable_fields = ("_offset",)

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        length: int,
        worker_config: WorkerConfig,
    ):
        """
        Create the epochized dataset.

        Args:
            dataset: The source dataset (possibly infinite)
            length: Number of samples to iterate before iteration stops (i.e. one epoch). When
                iteration continues, the original dataset iterator is resumed and does only restart
                if exhausted.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.length = length
        self._active_iter = None

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._offset = 0

    def __iter__(self) -> Iterator[T_sample]:
        # Compute the local length for this worker, i.e. all worker's lengths sum up to the total

        if self.worker_config.num_workers <= 1:
            local_length = self.length
        else:
            local_length = self.length // self.worker_config.num_workers
            if self.worker_config.rank_worker_id() < self.length % self.worker_config.num_workers:
                local_length += 1

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "EpochizeDataset.epoch_start",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "offset": self._offset,
                    "local_length": local_length,
                    "length": self.length,
                }
            )

        offset_range = list(range(self._offset, local_length))

        # Only iterate if there are samples to iterate
        if len(offset_range) > 0:
            if self._active_iter is None:
                self._active_iter = iter(self.dataset)

            for idx in offset_range:
                self._offset = (idx + 1) % local_length
                try:
                    sample = next(self._active_iter)
                except StopIteration:
                    break
                yield sample

        if self.worker_config.should_log(level=2):
            self.worker_config.worker_log(
                {
                    "t": "EpochizeDataset.epoch_end",
                    "r": self.worker_config.rank,
                    "w": self.worker_config.rank_worker_id(),
                    "offset": self._offset,
                    "local_length": local_length,
                    "length": self.length,
                }
            )

    def __len__(self) -> int:
        return self.length

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "length": self.length,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"EpochizeDataset(length={self.length}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import re
from typing import List, Optional

import click
import torch

from megatron.energon.epathlib import EPath
from megatron.energon.savable_loader import SavableDataLoaderState


def natural_sort_key(s):
    """
    Function to use for natural sorting of filenames.

    This splits the input string by numbers and non-numbers and ensures
    that numbers are compared as integers, not as strings.
    """
    return [int(text) if text.isdigit() else text.lower() for text in re.split(r"(\d+)", s)]


def detect_and_replicate_pattern(file_list):
    """
    Given a list of file paths, detect the single numeric pattern and return
    a function that, when called with integer n (starting from 0), generates
    the n-th filename following that pattern.

    Raises an Exception if no pattern or multiple patterns are found.
    """
    if not file_list:
        raise ValueError("Cannot detect a pattern from an empty list.")

    # -- 1) Sort the list using a natural key so that numbers compare numerically
    sorted_files = sorted(file_list, key=natural_sort_key)

    # -- 2) Tokenize each filename into [text, number, text, number, ...].
    #        We'll look for the pattern of tokens across all files.
    def tokenize_filename(fname):
        # Use the same split so that digit tokens are separated
        # from non-digit tokens.
        tokens = re.split(r"(\d+)", fname)
        # tokens is like ["f", "001", ".txt"] for "f001.txt"
        return tokens

    tokenized = [tokenize_filename(f) for f in sorted_files]

    # Check that all tokenized filenames have the same number of chunks:
    token_len = len(tokenized[0])
    for t in tokenized:
        if len(t) != token_len:
            raise Exception("Filenames do not share a consistent token structure.")

    # -- 3) Identify exactly one numeric token position that changes across all files.
    #        All other positions must be identical across the entire list.
    num_positions = []  # positions in the token list that differ
    for pos in range(token_len):
        # Check if this chunk is the same for all files or not:
        # We compare "raw text" for non-digit chunks, and "integer value" for digit chunks.

        # For the first file's token, check if it's digits or not
        example_token = tokenized[0][pos]
        example_is_digit = example_token.isdigit()

        # Collect how all files differ at this position
        all_tokens_at_pos = [t[pos] for t in tokenized]

        # If it's supposed to be a numeric token,
        # we compare the integer values to see if they differ or not.
        # If it's a non-numeric token, they must all be identical.
        if example_is_digit:
            # Parse integer values
            values = [int(x) if x.isdigit() else None for x in all_tokens_at_pos]
            # If *any* of them is None or they vary, we track that as "differences".
            # But let's see if indeed they differ across the files or not.
            if len(set(values)) > 1:
                # This token position changes among files
                num_positions.append(pos)
            else:
                # The numeric token is the same for all files, so no variation here
                pass
        else:
            # Non-digit token, must be identical across all files
            if len(set(all_tokens_at_pos)) != 1:
                raise Exception("Non-digit token differs among files. Invalid pattern.")

    # We expect exactly 1 changing numeric token position
    if len(num_positions) == 0:
        raise Exception("No numeric portion found that differs among files.")
    if len(num_positions) > 1:
        raise Exception("Multiple numeric portions found that differ. Not a single pattern.")

    varying_pos = num_positions[0]

    # -- 4) Extract the numeric values of that varying position for all sorted files,
    #        check consecutive increments and find the zero-padding width.
    numeric_values = [int(t[varying_pos]) for t in tokenized]

    # Check if consecutive differences are all +1
    for i in range(len(numeric_values) - 1):
        if numeric_values[i + 1] - numeric_values[i] != 1:
            raise Exception("Numeric values are not consecutive. Pattern is invalid.")

    # The "base" number is numeric_values[0], i.e. the value for n=0
    base_value = numeric_values[0]

    # The zero-padding width is based on the first file's numeric token
    zero_padding_width = len(tokenized[0][varying_pos])

    # -- 5) Construct the function that, given n, returns the enumerated filename.
    #        We'll verify it against the original sorted list as well.
    def generate_filename(n):
        # Rebuild the token array from the first file's tokens,
        # except we replace the one numeric token with (base_value + n) zero-padded.
        new_tokens = tokenized[0][:]

        new_int_value = base_value + n
        # zero-pad with the discovered width
        new_str_value = str(new_int_value).zfill(zero_padding_width)

        # Replace the numeric position
        new_tokens[varying_pos] = new_str_value

        # Join all tokens back into a string
        return "".join(new_tokens)

    # -- 6) Verify that generate_filename(i) reproduces the sorted list exactly
    #        for i in [0..len(sorted_files)-1].
    for i in range(len(sorted_files)):
        candidate = generate_filename(i)
        if candidate != sorted_files[i]:
            raise Exception(
                "Verification failed. The generated pattern does not match the input list."
            )

    # If we get here, everything is good. Return the generator function.
    return generate_filename


class RankStateIterable:
    """Iterates the SavableDatasetCheckpoints of mulitple ranks in a round-robin fashion."""

    def __init__(self, state_files: List[EPath]):
        state_file_names = [state_file.name for state_file in state_files]

        self.file_pattern_func = detect_and_replicate_pattern(state_file_names)
        self.num_states = len(state_files)

        # First open the first one to figure out if this is a global checkpoint or not
        first_state = torch.load(str(state_files[0]), weights_only=False)

        if isinstance(first_state, dict) and "dataloader_state_dict" in first_state:
            self.megatron_style = True
            first_state = first_state["dataloader_state_dict"]
        else:
            self.megatron_style = False

        if isinstance(first_state, SavableDataLoaderState):
            if self.megatron_style:
                self.rank_states = [first_state] + [
                    torch.load(str(state_file), weights_only=False)["dataloader_state_dict"]
                    for state_file in state_files[1:]
                ]
            else:
                self.rank_states = [first_state] + [
                    torch.load(str(state_file), weights_only=False)
                    for state_file in state_files[1:]
                ]
            self.is_global_checkpoint = False
        elif isinstance(first_state, list):
            assert len(state_files) == 1, "Global checkpoint must contain exactly one file"
            assert all(isinstance(state, SavableDataLoaderState) for state in first_state)
            self.rank_states = first_state
            self.is_global_checkpoint = True
        else:
            raise ValueError(f"Unknown checkpoint type: {type(first_state)}")

        self.rank_cur_worker = [0] * len(self.rank_states)
        self.rank_worker_offset = [state.next_worker_id for state in self.rank_states]

        self.rank_num_workers = [len(state.worker_states) for state in self.rank_states]
        assert all(
            self.rank_num_workers[0] == num_workers for num_workers in self.rank_num_workers
        ), "All ranks must have the same number of workers."

    def write_new_states_to_folder(
        self, output_folder: EPath, new_states: List[SavableDataLoaderState]
    ):
        for rank_idx, rank_state in enumerate(new_states):
            output_file = output_folder / self.file_pattern_func(rank_idx)
            if self.megatron_style:
                torch.save(
                    {"dataloader_state_dict": rank_state},
                    str(output_file),
                )
            else:
                torch.save(rank_state, str(output_file))

    def get_num_ranks(self):
        return len(self.rank_states)

    def get_num_workers(self):
        return self.rank_num_workers[0]

    def get_micro_batch_size(self):
        return self.rank_states[0].micro_batch_size

    def __iter__(self):
        """Iterates the SavableDatasetCheckpoints of mulitple ranks in a round-robin fashion."""
        for rank, state in enumerate(self.rank_states):
            for worker_state in state.worker_states:
                yield worker_state


@click.command(name="redist")
@click.argument(
    "input_files",
    nargs=-1,
    type=click.Path(file_okay=True, dir_okay=False, exists=True, path_type=EPath),
    required=True,
)
@click.argument(
    "output_path",
    type=click.Path(file_okay=False, dir_okay=True, path_type=EPath),
)
@click.option(
    "--new-world-size", type=int, help="Number of ranks to redistribute to", required=False
)
def command_redist(
    input_files: List[EPath], output_path: EPath, new_world_size: Optional[int] = None
):
    """Redistribute a checkpoint.

    Read checkpoint files from INPUT_FILES and redistribute them for a new
    number of ranks. Write the output to OUTPUT_PATH."""

    # Verify input files
    if not input_files:
        raise click.ClickException("No input files provided")

    input_file_list = sorted(input_files, key=lambda x: natural_sort_key(x.name))

    click.echo(f"Processing {len(input_file_list)} checkpoint files")

    # Determine if we're processing a single global checkpoint or multiple rank files
    rsi = RankStateIterable(input_file_list)

    if not rsi.rank_states:
        raise click.ClickException("No valid checkpoint states found")

    if new_world_size is None:
        click.echo(f"Current DP world size: {rsi.get_num_ranks()}")
        click.echo(f"Current number of workers per DP rank: {rsi.get_num_workers()}")
        new_world_size = click.prompt("Please enter the new DP world size", type=int)
        assert isinstance(new_world_size, int)

    if new_world_size <= 0:
        raise click.ClickException("New world size must be greater than 0")

    total_num_workers = rsi.get_num_workers() * rsi.get_num_ranks()
    assert total_num_workers % new_world_size == 0, (
        "New DP world size must be a multiple of the current DP world size"
    )
    new_workers_per_rank = total_num_workers // new_world_size

    # Ensure output directory exists
    output_path.mkdir(exist_ok=True, parents=True)

    new_rank_states = [list() for _ in range(new_world_size)]
    rsi_iter = iter(rsi)
    for rank_idx in range(new_world_size):
        for _ in range(new_workers_per_rank):
            state = next(rsi_iter)
            new_rank_states[rank_idx].append(state)

    assert all(
        len(new_rank_states[0]) == len(new_rank_states[rank]) for rank in range(1, new_world_size)
    ), "All ranks must have the same number of workers, also for the new distribution."

    new_states = [
        SavableDataLoaderState(
            worker_states=new_rank_state,
            next_worker_id=0,  # Reset the next worker ID
            micro_batch_size=rsi.get_micro_batch_size(),
        )
        for new_rank_state in new_rank_states
    ]

    # Save the redistributed checkpoint
    if rsi.is_global_checkpoint:
        # Save as a single global checkpoint file
        output_file = output_path / input_file_list[0].name
        torch.save(new_states, str(output_file))
        click.echo(f"Saved global checkpoint to {output_file}")
    else:
        rsi.write_new_states_to_folder(output_path, new_states)

        click.echo(f"Saved {new_world_size} rank checkpoint files to {output_path}")


@click.command(name="info")
@click.argument(
    "input_files",
    nargs=-1,
    type=click.Path(file_okay=True, dir_okay=False, exists=True, path_type=EPath),
    required=True,
)
def command_info(input_files: List[EPath]):
    """Display information about a checkpoint.

    Read a checkpoint from CHECKPOINT_PATH (either a single file or directory with *.pt files)
    and display information about it.
    """

    # Load the checkpoint(s)
    rsi = RankStateIterable(input_files)

    # Display basic information
    if rsi.is_global_checkpoint:
        click.echo("Checkpoint type: Global checkpoint")
    else:
        click.echo("Checkpoint type: Per-rank checkpoint files")

    click.echo(f"Number of DP ranks: {rsi.get_num_ranks()}")
    click.echo(f"Number of workers per DP rank: {rsi.get_num_workers()}")

    # Additional detailed information
    click.echo("\nDetailed information:")
    for rank_idx, state in enumerate(rsi.rank_states):
        if rsi.is_global_checkpoint:
            click.echo(f"  Rank {rank_idx}:")
        else:
            click.echo(f"  Rank {rank_idx} ({input_files[rank_idx].name}):")
        click.echo(f"    Next worker ID: {state.next_worker_id}")
        click.echo(f"    Number of worker states: {len(state.worker_states)}")


@click.group(
    name="checkpoint",
    context_settings=dict(help_option_names=["-h", "--help"]),
    invoke_without_command=True,
)
@click.pass_context
def command(ctx):
    """Tools for energon checkpoints."""

    # This is needed to show help if no subcommand is provided
    if ctx.invoked_subcommand is None:
        click.echo(command.get_help(ctx))


command.add_command(command_redist)
command.add_command(command_info)

if __name__ == "__main__":
    command()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import re
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Type, Union

from megatron.energon.cache import FileStore
from megatron.energon.cache.file_store import (
    SystemFileStore,
    WebdatasetFileStore,
)
from megatron.energon.dataset_config import load_config
from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.flavors import Sample
from megatron.energon.flavors.webdataset import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.metadata import check_dataset_info_present
from megatron.energon.metadataset.dataset_loader import DatasetLoader
from megatron.energon.metadataset.join_dataset_loader import JoinDatasetLoader, JoinedDatasetInfo
from megatron.energon.metadataset.loader_interface import (
    DatasetBlendMode,
    DatasetLoaderInterface,
    LoadedDatasetList,
)
from megatron.energon.metadataset.metadataset import Metadataset
from megatron.energon.worker import WorkerConfig

# Regex for any URL-like string (any protocol)
url_regex = re.compile(r"^(?P<protocol>[a-z][a-z0-9+.-]*)://(?P<path>.*)", re.IGNORECASE)


@edataclass
class AuxDatasetReference:
    path: Union[str, EPath]

    def post_initialize(self, mds_path: Optional[EPath] = None) -> None:
        assert mds_path is not None
        if not isinstance(self.path, EPath):
            self.path = mds_path.parent / self.path
        assert not self.path.is_file(), (
            "Auxiliary datasets must not be metadataset, but direct dataset references"
        )
        assert (self.path / MAIN_FOLDER_NAME / "index.sqlite").is_file(), (
            "Auxiliary datasets must be prepared Energon dataset"
        )

    def get_file_store(self) -> FileStore:
        assert isinstance(self.path, EPath), "Missing call to post_initialize"
        return WebdatasetFileStore(self.path)


@edataclass
class AuxFilesystemReference:
    fs_path: Union[str, EPath]

    def post_initialize(self, mds_path: Optional[EPath] = None) -> None:
        assert mds_path is not None
        if not isinstance(self.fs_path, EPath):
            self.fs_path = mds_path.parent / self.fs_path

    def get_file_store(self) -> FileStore:
        assert isinstance(self.fs_path, EPath), "Missing call to post_initialize"
        return SystemFileStore(self.fs_path)


@edataclass
class DatasetReference(DatasetLoaderInterface):
    path: Union[str, EPath]

    split_part: Optional[str] = None
    subflavors: Optional[Dict[str, Any]] = None
    shuffle_over_epochs_multiplier: Optional[int] = 1
    dataset_config: str = "dataset.yaml"
    split_config: str = "split.yaml"

    #: Auxiliary datasets. May only be specified for crude datasets for cooking. Cooking will get
    # these references to load data from. If specified as string, it will be interpreted as a
    # dataset path.
    aux: Optional[Dict[str, str]] = None

    _dataset: Optional[DatasetLoaderInterface] = None

    def post_initialize(self, mds_path: Optional[EPath] = None) -> None:
        assert mds_path is not None
        if not isinstance(self.path, EPath):
            self.path = mds_path.parent / self.path
        if self.path.is_file():
            assert self.aux is None, "Cannot specify auxiliary datasets for crude datasets"
            assert self.dataset_config == "dataset.yaml", "Must not set dataset_config"
            assert self.split_config == "split.yaml", "Must not set split_config"
            # Note: For backwards compatibility, the type must be Metadataset (V1).
            self._dataset = load_config(
                self.path,
                default_type=Metadataset,
                default_kwargs=dict(path=self.path),
            )
            self._dataset.post_initialize()
        elif check_dataset_info_present(self.path):
            self._dataset = DatasetLoader(path=self.path)
            self._dataset.post_initialize()
            if self.aux is not None:
                new_aux = {}
                for k, v in self.aux.items():
                    if m := url_regex.match(v):
                        if m.group("protocol") == "filesystem":
                            new_aux[k] = AuxFilesystemReference(fs_path=m.group("path"))
                        else:
                            raise ValueError(f"Unsupported protocol: {m.group('protocol')}")
                    else:
                        new_aux[k] = AuxDatasetReference(path=v)

                    new_aux[k].post_initialize(mds_path)
                self.aux = new_aux
        else:
            raise FileNotFoundError(self.path)

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        assert self._dataset is not None
        return self._dataset.prepare(split_part=split_part)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        if self.subflavors is not None:
            subflavors = {**self.subflavors, **(subflavors or {})}
        assert self._dataset is not None

        if shuffle_over_epochs_multiplier is None or self.shuffle_over_epochs_multiplier is None:
            # If no shuffling is requested, this has override priority.
            new_shuffle_over_epochs_multiplier = None
        elif shuffle_over_epochs_multiplier == -1 or self.shuffle_over_epochs_multiplier == -1:
            # Next priority is sampling without replacement.
            new_shuffle_over_epochs_multiplier = -1
        else:
            # Otherwise, multiply the shuffle over epochs multiplier.
            new_shuffle_over_epochs_multiplier = (
                shuffle_over_epochs_multiplier * self.shuffle_over_epochs_multiplier
            )
        result = self._dataset.get_datasets(
            training=training,
            split_part=self.split_part or split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=new_shuffle_over_epochs_multiplier,
            **kwargs,
        )
        if self.aux is not None:
            aux = {k: v.get_file_store() for k, v in self.aux.items()}
            for loaded_dataset in result.datasets:
                if loaded_dataset.aux is None:
                    loaded_dataset.aux = aux
                else:
                    loaded_dataset.aux.update(aux)
        return result


@edataclass
class JoinDatasetReference(DatasetReference):
    nonmatch: Literal["skip", "none", "error"] = "error"

    def post_initialize(self, mds_path: Optional[EPath] = None) -> DatasetLoader:
        assert mds_path is not None
        # Override and disable another metadataset reference, only allow direct dataset references.
        # Do not store the loader, the parent MetadatasetJoin will do that.
        if not isinstance(self.path, EPath):
            self.path = mds_path.parent / self.path
        if check_dataset_info_present(self.path):
            return DatasetLoader(
                path=self.path,
                split_part=self.split_part,
                subflavors=self.subflavors,
                shuffle_over_epochs_multiplier=self.shuffle_over_epochs_multiplier,
                dataset_config=self.dataset_config,
                split_config=self.split_config,
            )
        else:
            raise FileNotFoundError(self.path)

    def prepare(self, split_part: Optional[str] = None):
        assert False, (
            "JoinDatasetReference should not be used directly, but only by MetadatasetJoin"
        )

    def get_datasets(
        self,
        **kwargs,
    ) -> LoadedDatasetList:
        assert False, (
            "JoinDatasetReference should not be used directly, but only by MetadatasetJoin"
        )


@edataclass
class MetadatasetJoin(DatasetLoaderInterface):
    join: Union[List[JoinDatasetReference], Dict[str, JoinDatasetReference]]
    joiner: Union[Type[Sample], Callable[..., Sample]]

    split_part: Optional[str] = None
    subflavors: Optional[Dict[str, Any]] = None
    shuffle_over_epochs_multiplier: Optional[int] = 1
    dataset_config: str = "dataset.yaml"
    split_config: str = "split.yaml"

    _dataset: Optional[JoinDatasetLoader] = None

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is not None
        assert self.join is not None
        assert self.joiner is not None, "Must set joiner for joining datasets"
        assert self.dataset_config == "dataset.yaml", (
            "Cannot set dataset_config for joining datasets"
        )
        if isinstance(self.join, list):
            inner_loaders = [
                JoinedDatasetInfo(
                    dataset=join.post_initialize(mds_path),
                    nonmatch=join.nonmatch,
                )
                for join in self.join
            ]
        elif isinstance(self.join, dict):
            inner_loaders = {
                key: JoinedDatasetInfo(
                    dataset=join.post_initialize(mds_path),
                    nonmatch=join.nonmatch,
                )
                for key, join in self.join.items()
            }
        else:
            raise ValueError("Invalid join type")

        self._dataset = JoinDatasetLoader(
            datasets=inner_loaders,
            joiner=self.joiner,
            split_part=self.split_part,
            subflavors=self.subflavors,
            shuffle_over_epochs_multiplier=self.shuffle_over_epochs_multiplier,
            split_config=self.split_config,
        )
        self._dataset.post_initialize(mds_path)

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        assert self._dataset is not None, "Missing post_initialize call."
        return self._dataset.prepare(split_part=split_part)

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        assert self._dataset is not None, "Missing post_initialize call."
        return self._dataset.get_datasets(
            training=training,
            split_part=split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
            **kwargs,
        )


@dataclass
class BlendWeightMixin:
    weight: float = 1.0


@edataclass
class BlendDatasetReference(BlendWeightMixin, DatasetReference):
    pass


@edataclass
class BlendJoinDatasetReference(BlendWeightMixin, MetadatasetJoin):
    pass


@edataclass
class MetadatasetBlend(DatasetLoaderInterface):
    """Blending of datasets by specifying the sampling weight for the inner datasets."""

    blend: List[Union[BlendDatasetReference, BlendJoinDatasetReference]]

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is not None
        for dataset in self.blend:
            dataset.post_initialize(mds_path)

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        files = []
        for dataset in self.blend:
            files.extend(dataset.prepare(split_part=split_part))
        return files

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        sum_weight = sum(dataset.weight for dataset in self.blend)
        datasets = []
        for dataset in self.blend:
            inner_result = dataset.get_datasets(
                training=training,
                split_part=split_part,
                worker_config=worker_config,
                subflavors=subflavors,
                shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
                **kwargs,
            )
            if inner_result.blend_mode not in (
                DatasetBlendMode.NONE,
                DatasetBlendMode.DATASET_WEIGHT,
            ):
                raise ValueError(
                    "Can only blend datasets which are of the same blend mode. Cannot mix blend with blend_epochized."
                )
            for loaded_dataset in inner_result.datasets:
                if inner_result.blend_mode == DatasetBlendMode.DATASET_WEIGHT:
                    assert isinstance(loaded_dataset.weight, float)
                else:
                    assert inner_result.blend_mode == DatasetBlendMode.NONE
                    assert loaded_dataset.weight is None
                    assert loaded_dataset.repetitions is None
                    loaded_dataset.weight = 1.0
                loaded_dataset.weight = loaded_dataset.weight * dataset.weight / sum_weight
                datasets.append(loaded_dataset)
        return LoadedDatasetList(
            blend_mode=DatasetBlendMode.DATASET_WEIGHT,
            datasets=datasets,
        )


@dataclass
class BlendRepetitionsMixin:
    repetitions: Union[int, float] = 1


@edataclass
class BlendEpochizedDatasetReference(BlendRepetitionsMixin, DatasetReference):
    pass


@edataclass
class BlendEpochizedJoinDatasetReference(BlendRepetitionsMixin, MetadatasetJoin):
    pass


@edataclass
class MetadatasetBlendEpochized(DatasetLoaderInterface):
    """Blending of datasets, by specifying the number of repetitions for samples from the inner
    datasets. Ensures that the constraint, that samples are seen exactly this many times before
    repeating the "epoch" (i.e. one epoch contains the total number of repetitions for each inner
    dataset)."""

    blend_epochized: List[Union[BlendEpochizedDatasetReference, BlendEpochizedJoinDatasetReference]]

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is not None
        for dataset in self.blend_epochized:
            dataset.post_initialize(mds_path)

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        files = []
        for dataset in self.blend_epochized:
            files.extend(dataset.prepare(split_part=split_part))
        return files

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        datasets = []
        for dataset in self.blend_epochized:
            inner_result = dataset.get_datasets(
                training=training,
                split_part=split_part,
                worker_config=worker_config,
                subflavors=subflavors,
                shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
                **kwargs,
            )
            if inner_result.blend_mode not in (
                DatasetBlendMode.NONE,
                DatasetBlendMode.SAMPLE_REPETITIONS,
            ):
                raise ValueError(
                    "Can only blend datasets which are of the same blend mode. Cannot mix blend with blend_epochized."
                )
            for loaded_dataset in inner_result.datasets:
                if inner_result.blend_mode == DatasetBlendMode.SAMPLE_REPETITIONS:
                    assert isinstance(loaded_dataset.repetitions, (int, float))
                else:
                    assert loaded_dataset.weight is None
                    assert loaded_dataset.repetitions is None
                    loaded_dataset.repetitions = 1
                loaded_dataset.repetitions = dataset.repetitions * loaded_dataset.repetitions
                datasets.append(loaded_dataset)
        return LoadedDatasetList(
            blend_mode=DatasetBlendMode.SAMPLE_REPETITIONS,
            datasets=datasets,
        )


@edataclass
class MetadatasetV2(DatasetLoaderInterface):
    path: EPath
    splits: Dict[
        str, Union[MetadatasetBlend, MetadatasetBlendEpochized, MetadatasetJoin, DatasetReference]
    ]

    def post_initialize(self, mds_path: Optional[EPath] = None):
        assert mds_path is None
        for split in self.splits.values():
            split.post_initialize(self.path)

    def prepare(self, split_part: Optional[str] = None) -> Sequence[EPath]:
        # In the case of prepare for MetadatasetV2, we ignore the passed cache_path
        # and instead use the own path.
        # If someone runs energon prepare on a metadataset that refers to another metadataset,
        # any actions concerning the inner metadataset will be done on the inner metadataset's path.

        if split_part is None:
            files = []
            for split_part, split in self.splits.items():
                files.extend(split.prepare(split_part=split_part))
        else:
            files = self.splits[split_part].prepare(split_part=split_part)
        # Cleanup paths here
        cache_path = EPath(self.path.parent / f"{self.path.name}.cache")
        remove_files = set(cache_path.glob("*")) - set(files)
        for file in remove_files:
            print(f"Cleanup cache file {file}")
            file.unlink()
        # Do not return paths. This is a barrier
        return ()

    def get_datasets(
        self,
        *,
        training: bool,
        split_part: Union[Literal["train", "val", "test"], str],
        worker_config: WorkerConfig,
        subflavors: Optional[Dict[str, Any]] = None,
        shuffle_over_epochs_multiplier: Optional[int] = 1,
        **kwargs,
    ) -> LoadedDatasetList:
        return self.splits[split_part].get_datasets(
            training=training,
            split_part=split_part,
            worker_config=worker_config,
            subflavors=subflavors,
            shuffle_over_epochs_multiplier=shuffle_over_epochs_multiplier,
            **kwargs,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.wrappers.batch_dataset import BatchDataset
from megatron.energon.wrappers.blend_dataset import BlendDataset
from megatron.energon.wrappers.concat_dataset import ConcatDataset
from megatron.energon.wrappers.epochize_dataset import EpochizeDataset
from megatron.energon.wrappers.filter_dataset import FilterDataset
from megatron.energon.wrappers.gc_dataset import GcDataset
from megatron.energon.wrappers.group_batch_dataset import GroupBatchDataset
from megatron.energon.wrappers.iter_map_dataset import IterMapDataset
from megatron.energon.wrappers.limit_dataset import LimitDataset
from megatron.energon.wrappers.log_sample_dataset import LogSampleDataset
from megatron.energon.wrappers.map_dataset import MapDataset
from megatron.energon.wrappers.mix_batch_dataset import (
    MixBatchDataset,
    concat_pad,
    generic_concat,
    homogeneous_concat_mix,
)
from megatron.energon.wrappers.packing_dataset import PackingDataset
from megatron.energon.wrappers.repeat_dataset import RepeatDataset
from megatron.energon.wrappers.shuffle_buffer_dataset import ShuffleBufferDataset
from megatron.energon.wrappers.skip import SkipSample

__all__ = [
    "BatchDataset",
    "BlendDataset",
    "ConcatDataset",
    "EpochizeDataset",
    "FilterDataset",
    "GcDataset",
    "GroupBatchDataset",
    "IterMapDataset",
    "LimitDataset",
    "LogSampleDataset",
    "MapDataset",
    "MixBatchDataset",
    "RepeatDataset",
    "ShuffleBufferDataset",
    "SkipSample",
    "PackingDataset",
    "concat_pad",
    "generic_concat",
    "homogeneous_concat_mix",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import torch

from megatron.energon.transforms.mappers import TransformMapper


class CustomTransform(TransformMapper[torch.nn.Module]):
    """Abstract class for additional custom transforms.
    Inherit from this and override at least `apply_transform`.
    """



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from typing import Any, Dict, Generic, Iterator, Optional, TypeVar

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.watchdog import Watchdog
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


class WatchdogDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset wraps another dataset and watches the time it takes to yield samples."""

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        worker_config: WorkerConfig,
        timeout_seconds: Optional[float] = 60,
        initial_timeout_seconds: Optional[float] = None,
        fail_on_timeout: bool = False,
    ):
        """Construct the watchdog dataset, which wraps another dataset and watches
        the time it takes to yield samples from the wrapped dataset.

        Args:
            dataset: The input dataset to wrap
            worker_config: The worker configuration
            timeout_seconds: The timeout in seconds. If None, the watchdog is disabled.
            initial_timeout_seconds: The initial timeout in seconds. If None, the timeout is the same as timeout_seconds.
            fail_on_timeout: If True, stops the whole process upon timeout, after printing a stack trace.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.timeout_seconds = timeout_seconds
        self.initial_timeout_seconds = initial_timeout_seconds
        self.fail_on_timeout = fail_on_timeout

    def reset_state_own(self) -> None:
        pass

    def __len__(self):
        return len(self.dataset)

    def _watchdog_trigger(self) -> None:
        if self.fail_on_timeout:
            # Raising an exception here will kill the whole process
            raise TimeoutError(
                f"Watchdog triggered. Sample processing took longer than {self.timeout_seconds} seconds."
            )
        else:
            warnings.warn(
                f"Watchdog triggered. Sample processing took longer than {self.timeout_seconds} seconds.",
                RuntimeWarning,
            )

    def __iter__(self) -> Iterator[T_sample]:
        if self.timeout_seconds is None:
            yield from self.dataset
        else:
            watchdog = Watchdog(
                timeout=self.timeout_seconds,
                initial_timeout=self.initial_timeout_seconds,
                callback=self._watchdog_trigger,
                enabled=False,
            )
            yield from watchdog.watch_iter(self.dataset)

    def config(self) -> Dict[str, Any]:
        # Watchdog is transparent, it won't change the samples
        return self.dataset.config()

    def __str__(self):
        return f"WatchdogDataset(dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
from dataclasses import dataclass


@dataclass
class KeyframeInfo:
    """
    Information about a detected keyframe

    The exact meaning of the information will vary per container, however in general

    index: the unit of representation for a frame (e.g., frame number)
    pts: a timestamp that can be used by a decoder to seek to exactly this frame

    """

    index: int
    pts: int

    def __eq__(self, other) -> bool:
        if isinstance(other, KeyframeInfo):
            return self.index == other.index

        return self.index == other

    def __lt__(self, other) -> bool:
        if isinstance(other, KeyframeInfo):
            return self.index < other.index

        return self.index < other



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.flavors.base_dataset import (
    BaseCoreDatasetFactory,
    PinMemoryMixin,
    Sample,
    SavableDataset,
)
from megatron.energon.flavors.captioning import CaptioningSample, CaptioningWebdataset
from megatron.energon.flavors.crude import CrudeSample, CrudeWebdataset
from megatron.energon.flavors.image import ImageSample, ImageWebdataset
from megatron.energon.flavors.image_classification import (
    ImageClassificationSample,
    ImageClassificationWebdataset,
)
from megatron.energon.flavors.interleaved import InterleavedSample, InterleavedWebdataset
from megatron.energon.flavors.multichoice_vqa import MultiChoiceVQASample, MultiChoiceVQAWebdataset
from megatron.energon.flavors.ocr import OCRSample, OCRWebdataset
from megatron.energon.flavors.similarity_interleaved import (
    SimilarityInterleavedSample,
    SimilarityInterleavedWebdataset,
)
from megatron.energon.flavors.text import TextSample, TextWebdataset
from megatron.energon.flavors.vid_qa import VidQASample, VidQAWebdataset
from megatron.energon.flavors.vqa import VQASample, VQAWebdataset
from megatron.energon.flavors.vqa_and_ocr import VQAOCRWebdataset
from megatron.energon.flavors.webdataset import (
    AVData,
    AVDecoder,
    AVDecoderType,
    BaseWebdatasetFactory,
    DefaultDecoderWebdatasetFactory,
    DefaultGenericWebdatasetFactory,
    EmptyDatasetError,
    ImageDecoderType,
    JoinedWebdatasetFactory,
    SampleDecoder,
    StandardWebdatasetFactory,
    WebdatasetMeta,
)

__all__ = [
    "AVData",
    "AVDecoder",
    "AVDecoderType",
    "BaseCoreDatasetFactory",
    "BaseWebdatasetFactory",
    "CaptioningSample",
    "CaptioningWebdataset",
    "CrudeSample",
    "CrudeWebdataset",
    "DefaultDecoderWebdatasetFactory",
    "DefaultGenericWebdatasetFactory",
    "EmptyDatasetError",
    "ImageClassificationSample",
    "ImageClassificationWebdataset",
    "ImageDecoderType",
    "ImageSample",
    "ImageWebdataset",
    "InterleavedSample",
    "InterleavedWebdataset",
    "JoinedWebdatasetFactory",
    "MultiChoiceVQASample",
    "MultiChoiceVQAWebdataset",
    "OCRSample",
    "OCRWebdataset",
    "PinMemoryMixin",
    "Sample",
    "SampleDecoder",
    "SavableDataset",
    "SimilarityInterleavedSample",
    "SimilarityInterleavedWebdataset",
    "StandardWebdatasetFactory",
    "TextSample",
    "TextWebdataset",
    "VidQASample",
    "VidQAWebdataset",
    "VQAOCRWebdataset",
    "VQASample",
    "VQAWebdataset",
    "WebdatasetMeta",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import logging
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, TypeVar

import numpy as np

from megatron.energon.cache import FileStore
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import (
    BaseCoreDatasetFactory,
    SavableDataset,
    legacy_handler,
)
from megatron.energon.flavors.webdataset.error_handler import ErrorHandler
from megatron.energon.flavors.webdataset.metadata import WebdatasetMeta
from megatron.energon.flavors.webdataset.prepare import WebdatasetPreparator
from megatron.energon.flavors.webdataset.sample_loader import (
    RawSampleData,
    WebdatasetSampleLoaderDataset,
)
from megatron.energon.flavors.webdataset.sharder import Sharder
from megatron.energon.flavors.webdataset.structs import FilteredSample, ShardInfo, reraise_exception
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.map_dataset import MapDataset

T_sample = TypeVar("T_sample", covariant=True)
T = TypeVar("T", covariant=True)

logger = logging.getLogger(__name__)


class BaseWebdatasetFactory(
    BaseCoreDatasetFactory[T_sample],
    WebdatasetPreparator,
    Sharder,
    ErrorHandler,
    Generic[T_sample],
    ABC,
):
    """
    Base class for all webdataset sample loader factories. Applies proper sharding across workers.
    """

    path: EPath

    training: bool
    worker_config: WorkerConfig

    shards: List[ShardInfo]
    rank_shards: List[List[Sequence[ShardInfo]]]

    def __init__(
        self,
        path: EPath,
        *,
        split_part: str,
        training: bool,
        worker_config: WorkerConfig,
        shuffle_over_epochs: Optional[int] = 1,
        parallel_shard_iters: Optional[int] = None,
        max_samples_per_sequence: Optional[int] = None,
        split_config: str = "split.yaml",
        part_filter: Optional[Callable[[str], bool]] = None,
        handler: Callable[
            [Exception, Optional[str], Optional[list[SourceInfo]]], None
        ] = reraise_exception,
    ):
        """
        Base factory for the webdataset sample loader.

        Args:
            path: Path to the dataset.
            split_part: Which part to load (e.g. 'train', 'val', 'test').
            training: If true, apply shuffling and loop the dataset.
            worker_config: Configuration for the workers.
            shuffle_over_epochs: Only effective if training=True.
                How many epochs to shuffle over if training.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: Number of parallel opened shards per worker, shuffling between.
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                    will be sequentially iterated).
            split_config: Config file to use for shard split definitions.
            part_filter: (internal) Function for filtering tar files by dict keys
            handler: Exception handler. Args: (exception, key).
        """
        assert self.__sample_type__ is not None, f"Class {type(self)} must define __sample_type__"
        wds_meta = WebdatasetMeta.from_config(
            path=path, split_part=split_part, split_config=split_config
        )
        self.path = path
        self.paths = [path]
        self.shards = wds_meta.shards
        self.sample_excludes = wds_meta.sample_excludes
        self.split_part_files = wds_meta.split_part_files
        self.training = training
        self.worker_config = worker_config
        self.shuffle_over_epochs = shuffle_over_epochs
        self.parallel_shard_iters = parallel_shard_iters
        self.max_samples_per_sequence = max_samples_per_sequence
        self.part_filter = part_filter
        self.handler = legacy_handler(handler)

    def __len__(self) -> int:
        return sum(shard.count for shard in self.shards)

    def build(self, worker_rotation_offset: int = 0) -> SavableDataset[T_sample]:
        from megatron.energon.flavors.webdataset.itar_reader import ShardInfosITarReader

        if self.parallel_shard_iters is None:
            if self.training:
                # 16 seems to be a good choice since we don't want too many file handles open
                parallel_shard_iters = 16
            else:
                parallel_shard_iters = 1
        else:
            parallel_shard_iters = self.parallel_shard_iters

        workers_sample_slice_offsets = self.shard_workers(
            self.shards,
            worker_config=self.worker_config,
            max_samples_per_sequence=self.max_samples_per_sequence,
            rotation_offset=worker_rotation_offset,
        )
        _print_shard_slices(self.worker_config, self.shards, workers_sample_slice_offsets)

        itar_reader = ShardInfosITarReader(
            self.path,
            self.shards,
            part_filter=self.part_filter,
            sample_filter=self.sample_filter,
            itar_cache_size=parallel_shard_iters,
        )

        dataset = WebdatasetSampleLoaderDataset(
            join_readers=[itar_reader],
            workers_sample_slice_offsets=workers_sample_slice_offsets,
            worker_config=self.worker_config,
            shuffle_over_epochs=self.shuffle_over_epochs if self.training else None,
            parallel_slice_iters=parallel_shard_iters,
        )
        return MapDataset(
            dataset,
            self._load_sample_raw,
            error_handler=self.error_handler,
            stateless_map_fn=True,
            map_fn_config=self.config,
            worker_config=self.worker_config,
        )

    def as_file_store(self) -> "FileStore":
        from megatron.energon.cache.file_store import WebdatasetFileStore

        return WebdatasetFileStore(self.path)

    def sample_filter(self, key: str) -> bool:
        return key not in self.sample_excludes

    def _load_sample_raw(self, raw_sample: RawSampleData) -> T_sample:
        # Just a wrapper for the inner tuple. Tuple should be of length 1.
        assert len(raw_sample.data) == 1 and raw_sample.data[0] is not None
        return self.load_sample(raw_sample.data[0])

    @abstractmethod
    def load_sample(self, raw_data: FilteredSample) -> T_sample:
        """Loads the sample from the dataset."""
        ...

    def config(self) -> Dict[str, Any]:
        return dict(
            type=type(self).__qualname__,
            training=self.training,
            _path=str(self.path),
            shards=[
                dict(
                    name=shard.name,
                    count=shard.count,
                    _path=str(shard.path),
                )
                for shard in self.shards
            ],
            sample_excludes=list(self.sample_excludes),
            shuffle_over_epochs=self.shuffle_over_epochs,
            parallel_shard_iters=self.parallel_shard_iters,
            max_samples_per_sequence=self.max_samples_per_sequence,
        )

    def __str__(self):
        return f"{type(self).__name__}(path={self.path})"


def _print_shard_slices(
    worker_config: WorkerConfig, shards: List[ShardInfo], slice_offsets: Sequence[Sequence[int]]
):
    shard_starts = np.cumsum([0] + [shard.count for shard in shards])

    def shard_range_info(start: int, end: int) -> str:
        start_shard_idx = np.searchsorted(shard_starts, start, side="right") - 1
        end_shard_idx = np.searchsorted(shard_starts, end, side="left") - 1
        if start_shard_idx == end_shard_idx:
            shard = shards[start_shard_idx]
            if start - shard_starts[start_shard_idx] == 0:
                start_str = "(start)"
            else:
                start_str = ""
            if end - shard_starts[start_shard_idx] == shard.count:
                end_str = "(end)"
            else:
                end_str = ""
            return f"{shard.name}[{start - shard_starts[start_shard_idx]}{start_str}, {end - shard_starts[start_shard_idx]}{end_str}]"
        else:
            start_shard = shards[start_shard_idx]
            end_shard = shards[end_shard_idx]
            if start - shard_starts[start_shard_idx] == 0:
                start_str = "(start)"
            else:
                start_str = ""
            if end - shard_starts[end_shard_idx] == end_shard.count:
                end_str = "(end)"
            else:
                end_str = ""
            return f"{start_shard.name}[{start - shard_starts[start_shard_idx]}{start_str},]-{end_shard.name}[,{end - shard_starts[end_shard_idx]}{end_str}]"

    for worker_idx, sample_slice_offsets in enumerate(slice_offsets):
        start_idx = sample_slice_offsets[0]
        end_idx = sample_slice_offsets[-1]

        if len(sample_slice_offsets) > 6:
            offset_str = f"{', '.join(str(o) for o in sample_slice_offsets[:3])} ...<{len(sample_slice_offsets) - 6}> {', '.join(str(o) for o in sample_slice_offsets[-3:])}"
        else:
            offset_str = ", ".join(str(o) for o in sample_slice_offsets)
        if len(sample_slice_offsets) > 6:
            slices_str = (
                ", ".join(
                    shard_range_info(start, end)
                    for start, end in zip(sample_slice_offsets[:3], sample_slice_offsets[1:4])
                )
                + f" ...<{len(sample_slice_offsets) - 6}> "
                + ", ".join(
                    shard_range_info(start, end)
                    for start, end in zip(sample_slice_offsets[-4:-1], sample_slice_offsets[-3:])
                )
            )
        else:
            slices_str = ", ".join(
                shard_range_info(start, end)
                for start, end in zip(sample_slice_offsets[:-1], sample_slice_offsets[1:])
            )

        print(
            f"rank={worker_config.rank}, worker={worker_idx}: sample_range=[{start_idx}, {end_idx}] in {len(sample_slice_offsets) - 1} slices, "
            f"sum(count)={end_idx - start_idx}: indexes=[{offset_str}] slices=[{slices_str}]"
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABC
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from megatron.energon.cache import FileStore
from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import (
    BaseCoreDatasetFactory,
    Sample,
    SavableDataset,
    legacy_handler,
    set_sample_restore_key,
)
from megatron.energon.flavors.webdataset.base_webdataset import BaseWebdatasetFactory
from megatron.energon.flavors.webdataset.error_handler import ErrorHandler
from megatron.energon.flavors.webdataset.indexing import JoinIndexReader
from megatron.energon.flavors.webdataset.itar_reader import JoinIndexFileITarReader
from megatron.energon.flavors.webdataset.sample_loader import (
    RawSampleData,
    WebdatasetSampleLoaderDataset,
)
from megatron.energon.flavors.webdataset.sharder import Sharder
from megatron.energon.flavors.webdataset.structs import ShardInfo, reraise_exception
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.map_dataset import MapDataset

T_sample = TypeVar("T_sample", covariant=True)


class JoinedWebdatasetFactory(
    BaseCoreDatasetFactory[T_sample], Sharder, ErrorHandler[T_sample], Generic[T_sample], ABC
):
    """
    Base class for all webdataset loaders. Applies proper sharding across workers. Can join multiple datasets.
    """

    training: bool
    worker_config: WorkerConfig
    shuffle_over_epochs: Optional[int] = 1
    parallel_shard_iters: Optional[int]
    max_samples_per_sequence: Optional[int]
    join_index: EPath
    handler: Callable[[Exception, Optional[str], Optional[list[SourceInfo]]], None]

    shards: List[Sequence[ShardInfo]]
    part_datasets: SavableDataset[T_sample]

    inner_datasets: List[BaseWebdatasetFactory]
    inner_dataset_keys: Optional[List[str]]
    _sample_joiner: Callable[..., T_sample]

    def __init__(
        self,
        inner_datasets: Union[Sequence[BaseWebdatasetFactory], Mapping[str, BaseWebdatasetFactory]],
        *,
        training: bool,
        worker_config: WorkerConfig,
        shuffle_over_epochs: Optional[int] = 1,
        parallel_shard_iters: Optional[int] = None,
        max_samples_per_sequence: Optional[int] = None,
        join_index: EPath,
        joiner: Union[Type[T_sample], Callable[..., T_sample]],
        handler: Callable[
            [Exception, Optional[str], Optional[list[SourceInfo]]], None
        ] = reraise_exception,
    ):
        """
        Constructs the loader for a joined webdataset. The samples from the inner datasets are joined into a single
        sample using the joiner function.

        Args:
            inner_dataset: The inner datasets. Must be loaded internally with `_is_composed=True`.
                Either a list (\\*args for joiner) or a dict (\\*\\*kwargs for joiner) of datasets,
                where the samples will be passed to the joiner function as \\*args or \\*\\*kwargs.
            training: If true, apply shuffling and loop the dataset.
            worker_config: Configuration for the workers.
            shuffle_over_epochs: Only effective if training=True.
                How many epochs to shuffle over if training.
                If = 1, every sample is seen exactly once per epoch.
                If > 1, samples (or rather shard slices) are shuffled within this number of epochs
                (i.e. randomly selected without replacement).
                If -1, the shards are effectively shuffle over infinite epochs (i.e. shard slices
                are drawn with replacement).
            parallel_shard_iters: Number of parallel opened shards per worker, shuffling between.
            max_samples_per_sequence: Maximum number of samples per sequence (=how many samples
                    will be sequentially iterated).
            join_index: Path to the join index file. Only required for join_method="left".
            joiner: Type of the joined samples or a method for joining the samples.
            handler: Exception handler. Args: (exception, key).
        """
        self.__sample_type__ = joiner
        assert all(not hasattr(d, "dataset") for d in inner_datasets), (
            "Inner dataset was not instantiated with _is_composed=True"
        )
        if isinstance(joiner, type) and issubclass(joiner, Sample):
            joiner = joiner.from_joined
        else:
            assert callable(joiner), f"Joiner {joiner} must be a callable or a Sample subclass"
        if isinstance(inner_datasets, Mapping):
            inner_keys = list(inner_datasets.keys())
            self.inner_dataset_keys = inner_keys
            # Wrap the joiner to pass the samples as kwargs
            self._sample_joiner = lambda *samples: joiner(**dict(zip(inner_keys, samples)))
            inner_datasets = list(inner_datasets.values())
        else:
            assert isinstance(inner_datasets, Sequence)
            self._sample_joiner = joiner
            self.inner_dataset_keys = None

        self.join_index = join_index
        self.inner_datasets = inner_datasets
        self.shards = list(zip(*(dataset.shards for dataset in self.inner_datasets)))
        self.training = training
        self.worker_config = worker_config
        self.shuffle_over_epochs = shuffle_over_epochs
        self.parallel_shard_iters = parallel_shard_iters
        self.max_samples_per_sequence = max_samples_per_sequence
        self.handler = legacy_handler(handler)

    def __len__(self) -> int:
        return sum(shard.count for shard in self.inner_datasets[0].shards)

    def build(self, worker_rotation_offset: int = 0) -> SavableDataset[T_sample]:
        if self.parallel_shard_iters is None:
            if self.training:
                # 16 seems to be a good choice since we don't want too many file handles open
                parallel_shard_iters = 16
            else:
                parallel_shard_iters = 1
        else:
            parallel_shard_iters = self.parallel_shard_iters

        # Get join index, get size, distribute samples
        # Get samples for each worker on current rank
        assert self.join_index.is_file(), (
            f"Join index {self.join_index} does not exist, did you prepare the metadataset? "
            "If you already prepared the metadataset, the join index might be outdated due to "
            "modifications to the inner datasets. In this case, you need to re-prepare the metadataset."
        )

        with JoinIndexReader(self.join_index) as jir:
            total_samples = len(jir)

        workers_sample_slice_offsets = self.slice_workers(
            total_samples,
            worker_config=self.worker_config,
            max_samples_per_sequence=self.max_samples_per_sequence,
            rotation_offset=worker_rotation_offset,
        )

        for worker_idx, sample_slice_offsets in enumerate(workers_sample_slice_offsets):
            start_idx = sample_slice_offsets[0]
            end_idx = sample_slice_offsets[-1]

            if len(sample_slice_offsets) > 6:
                offset_str = f"{', '.join(str(o) for o in sample_slice_offsets[:3])} ...<{len(sample_slice_offsets) - 6}> {', '.join(str(o) for o in sample_slice_offsets[-3:])}"
            else:
                offset_str = ", ".join(str(o) for o in sample_slice_offsets)

            print(
                f"rank={self.worker_config.rank}, worker={worker_idx}: sample_range=[{start_idx}, {end_idx}) in {len(sample_slice_offsets) - 1} slices, "
                f"sum(count)={end_idx - start_idx}: [{offset_str}]"
            )

        itar_readers = [
            JoinIndexFileITarReader(
                index_file=self.join_index,
                column=col_idx,
                tar_filenames=indexed_dataset.split_part_files,
                base_path=indexed_dataset.path,
                part_filter=indexed_dataset.part_filter,
                itar_cache_size=parallel_shard_iters,
            )
            for col_idx, indexed_dataset in enumerate(self.inner_datasets)
        ]

        dataset = WebdatasetSampleLoaderDataset(
            join_readers=itar_readers,
            workers_sample_slice_offsets=workers_sample_slice_offsets,
            worker_config=self.worker_config,
            shuffle_over_epochs=self.shuffle_over_epochs if self.training else None,
            parallel_slice_iters=parallel_shard_iters,
        )
        return self._process_samples(dataset)

    def as_file_store(self) -> FileStore:
        raise NotImplementedError("Not supported on joined datasets")

    @property
    def paths(self) -> List[EPath]:
        return [dataset.path for dataset in self.inner_datasets]

    def _process_samples(self, dataset: SavableDataset[RawSampleData]) -> SavableDataset[T_sample]:
        """Internally loads the sample."""
        return MapDataset(
            dataset,
            self.load_sample,
            error_handler=self.error_handler,
            stateless_map_fn=True,
            map_fn_config=self.config,
            worker_config=self.worker_config,
        )

    def load_sample(self, samples: RawSampleData) -> T_sample:
        assert len(samples.data) > 0 and samples.data[0] is not None, "Always need primary sample"
        # First call the loaders of all inner datasets
        loaded_samples = tuple(
            None if sample is None else dataset.load_sample(sample)
            for dataset, sample in zip(self.inner_datasets, samples.data)
        )
        # Then combine the loaded smaples into the final type
        return set_sample_restore_key(
            self._sample_joiner(*loaded_samples),
            *samples.__restore_key__,
            src=self,
            fail_otherwise=True,
        )

    def config(self) -> Dict[str, Any]:
        return dict(
            type=type(self).__qualname__,
            joined_datasets=[dataset.config() for dataset in self.inner_datasets],
            training=self.training,
            shuffle_over_epochs=self.shuffle_over_epochs,
            parallel_shard_iters=self.parallel_shard_iters,
            max_samples_per_sequence=self.max_samples_per_sequence,
        )

    def __str__(self):
        return f"{type(self).__name__}(paths={self.paths})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import field
from typing import Dict, List, Optional, Tuple, TypedDict

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.source_info import SourceInfo


@edataclass
class WebdatasetInfo:
    """Info about a webdataset. Format for `.nv-meta/.info.yaml` or `.nv-meta/.info.json`."""

    #: The version of the energon library that was used to prepare the dataset
    energon_version: Optional[str] = None
    #: Maps shard name to number of samples in that shard
    shard_counts: Dict[str, int]


@edataclass
class WebdatasetSplits:
    """Info about the splits of a webdataset. Format for `.nv-meta/split.yaml` or `.nv-meta/split.json`
    (or custom user yaml/json)."""

    #: Maps split part to list of shard names
    split_parts: Dict[str, List[str]]
    #: Set of "<shard name>" or "<shard name>/<sample index>" to exclude
    exclude: List[str] = field(default_factory=list)


@edataclass
class ShardInfo:
    """Info about a single shard as passed through internally. Not exposed to the user."""

    #: Name of the shard file (relative path from the nvinfo dir)
    name: str
    #: The path to the shard file
    path: EPath
    #: The number of samples in this shard
    count: int


class FilteredSample(TypedDict):
    """This is just a definition for the internal loaders. Not exposed to the user."""

    #: The key of the sample within the tar file.
    #: If the tar file contains files 12.jpg and 12.txt,
    #: those two files make one sample with the key "12"
    __key__: str
    #: The base name of the shard file e.g. "shard_000"
    __shard__: str
    #: Globally unique key to restore a sample from disk.
    #: For example `("Webdataset", 123)` would restore the sample at index 123.
    __restore_key__: Tuple[str, int]
    #: The source information for the sample.
    __sources__: tuple[SourceInfo, ...]


def reraise_exception(
    exc: Exception, key: Optional[str], sources: Optional[list[SourceInfo]] = None
) -> None:
    if sources:
        raise Exception(
            f"For sample {key!r} from {', '.join(f'{source.dataset_path}[{source.index}] {source.shard_name}{source.file_names!r}' for source in sources)}"
        ) from exc
    elif key:
        raise Exception(f"For sample {key!r}") from exc
    else:
        raise



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path
from typing import Any, Dict, Optional, Type, TypeVar, Union

from megatron.energon.epathlib import EPath
from megatron.energon.eyaml import load_yaml
from megatron.energon.flavors import BaseCoreDatasetFactory, StandardWebdatasetFactory
from megatron.energon.flavors.webdataset import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.metadata import check_dataset_info_present
from megatron.energon.typed_converter import JsonParser
from megatron.energon.worker import WorkerConfig

T = TypeVar("T")


def load_config(
    path: Union[EPath, Dict[str, Any]],
    *,
    default_type: Type[T],
    default_kwargs: Optional[Dict[str, Any]] = None,
    parser: JsonParser = JsonParser(strict=True),
) -> T:
    """
    Loads a config from a file or directly from a dictionary.

    Args:
        path: Path to the config to load or a dictionary containing the config.
        default_type: If set, this is the type to use if no type is specified in the config.
        default_kwargs: Default kwargs to use, will be overridden by the config.

    Returns:
        The instantiated type.
    """
    if isinstance(path, dict):
        data = path
    else:
        # Read the config from a file
        with path.open("rb") as f:
            data: dict = load_yaml(f)

    if default_kwargs is not None:
        new_data = default_kwargs.copy()
        new_data.update(data)
        data = new_data

    return parser.raw_to_instance(data, default_type)


T_sample = TypeVar("T_sample", covariant=True)


def get_dataset_from_config(
    path: Union[EPath, Path, str],
    *,
    dataset_config: str = "dataset.yaml",
    split_config: str = "split.yaml",
    split_part: str = "train",
    training: bool = True,
    subflavors: Optional[Dict[str, Any]] = None,
    worker_config: WorkerConfig,
    sample_type: Optional[Type[T_sample]] = None,
    **kwargs,
) -> BaseCoreDatasetFactory[T_sample]:
    """
    Gets a dataset from a config path.

    Args:
        path: Path to the folder where the `.nv-meta` folder is contained.
        dataset_config: Filename of the dataset config file (`path / '.nv-meta' / config`)
        split_config: Filename of the split config file (`path / '.nv-meta' / split_config`)
        split_part: Name of the split to load.
        training: If true, apply training randomization and loop the dataset.
        subflavors: Merge-Override the __subflavors__ property of each sample.
        worker_config: If set, use this worker config instead of the default one.
        sample_type: Type of the samples to load, only used to ensure typing.
        **kwargs: Additional arguments to be passed to the dataset constructor.

    Returns:
        The instantiated dataset
    """
    path = EPath(path)
    if not check_dataset_info_present(path):
        raise ValueError(
            f"Path {path} does not contain a {MAIN_FOLDER_NAME}/.info.yaml or .info.json file. Did you forget to "
            f"prepare the dataset? Please check the documentation for an introduction to dataset "
            f"preparation."
        )
    dataset: BaseCoreDatasetFactory[T_sample] = load_config(
        path / MAIN_FOLDER_NAME / dataset_config,
        default_kwargs=dict(
            path=path,
            split_config=split_config,
            split_part=split_part,
            training=training,
            worker_config=worker_config,
            **kwargs,
        ),
        default_type=StandardWebdatasetFactory,
    )
    if subflavors is not None:
        dataset.subflavors.update(subflavors)
    if sample_type is not None:
        assert issubclass(dataset.__sample_type__, sample_type), (
            f"Sample of type {dataset.__sample_type__} is not a subclass of {sample_type}."
        )
    return dataset



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union

from megatron.energon.epathlib import EPath
from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.flavors.webdataset.base_webdataset import BaseWebdatasetFactory
from megatron.energon.flavors.webdataset.config import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.field_access import field_access, split_field_access
from megatron.energon.flavors.webdataset.structs import FilteredSample
from megatron.energon.module_loader import ModuleLoader

T_sample = TypeVar("T_sample", covariant=True)


class DefaultGenericWebdatasetFactory(BaseWebdatasetFactory[T_sample], Generic[T_sample]):
    """
    Default implementation of webdataset for generic samples and the generic config interface for use with dataset.yaml.
    """

    _sample_loader: Callable[[Dict[str, Any]], Dict[str, Any]]

    def __init__(
        self,
        path: EPath,
        *,
        subflavors: Optional[Dict[str, Any]] = None,
        field_map: Optional[Dict[str, str]] = None,
        sample_loader: Optional[Union[str, Callable[[dict], dict]]] = None,
        part_filter: Optional[Union[str, List[str], Callable[[str], bool]]] = None,
        **kwargs,
    ):
        """
        Factory for the webdataset sample loader and basic configuration options.

        Args:
            subflavors: Subflavors dictionary to set for all loaded samples.
            field_map: Mapping from the webdataset fields to the sample fields.
            sample_loader: Function to load the sample from the webdataset fields. May be a string
                in order to load a function from a module, or a callable directly.
            part_filter: Filter for the parts to load. May be a string in order to load a function
                from a module, or a callable directly.
            **kwargs: Args passed to parent constructor.
        """
        assert (field_map is None) != (sample_loader is None), (
            "Either field_map or sample_loader must be provided."
        )
        if sample_loader is not None:
            assert part_filter is not None, (
                "part_filter must be provided if sample_loader is provided."
            )
            module_loader = ModuleLoader()
            if isinstance(sample_loader, str):
                sample_loader = module_loader.get_function(
                    sample_loader, "sample_loader", relative_path=path / MAIN_FOLDER_NAME
                )
            else:
                assert callable(sample_loader)
                sample_loader = sample_loader
            if isinstance(part_filter, list):
                parts = set(part_filter)
                part_filter = lambda part: part in parts
            elif isinstance(part_filter, str):
                part_filter = module_loader.get_function(
                    part_filter, "part_filter", relative_path=path / MAIN_FOLDER_NAME
                )
            else:
                assert callable(part_filter)
            self._sample_loader = sample_loader
        else:
            assert field_map is not None
            assert part_filter is None
            # Split field map fields by json[field][field]
            fields = {key: split_field_access(field) for key, field in field_map.items()}
            assert set(field.name for field in dataclasses.fields(self.__sample_type__)).issuperset(
                fields.keys()
            ) and set(
                field.name
                for field in dataclasses.fields(self.__sample_type__)
                if field.default is not dataclasses.MISSING
                and field.default_factory is not dataclasses.MISSING
            ).issubset(field_map.keys()), (
                f"field_map does not map to type {self.__sample_type__.__name__} fields"
            )
            self._sample_loader = lambda sample: {
                k: field_access(sample, v) for k, v in fields.items()
            }
            parts = set(access[0] for options in fields.values() for access in options)
            part_filter = lambda part: part in parts
        inner_sample_loader = self._sample_loader
        self._sample_loader = lambda sample: {
            "__key__": sample["__key__"],
            **inner_sample_loader(sample),
            "__restore_key__": sample["__restore_key__"],
            "__subflavors__": self.subflavors,
            "__sources__": sample["__sources__"],
        }
        super().__init__(path, **kwargs, part_filter=part_filter)
        self.subflavors = subflavors or {}

    def load_sample(self, sample: FilteredSample) -> T_sample:
        return self.__sample_type__(**self._sample_loader(sample))

    def config(self) -> Dict[str, Any]:
        return dict(
            **super().config(),
            subflavors=self.subflavors,
            sample_loader=SavableDataset._function_config(self._sample_loader),
        )



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Generic, Iterator, Tuple, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.rng import WorkerRng
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset
from megatron.energon.wrappers.buffer import SavableSampleBuffer

T_sample = TypeVar("T_sample")


class ShuffleBufferDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """Shuffle buffer for the dataset."""

    size: int
    _worker_rng: WorkerRng
    _active_buffer: SavableSampleBuffer[T_sample]

    _savable_fields = ("_active_buffer", "_worker_rng")

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        size: int,
        *,
        worker_config: WorkerConfig,
    ):
        """Create a shuffle buffer for the dataset."""
        super().__init__(dataset, worker_config=worker_config)
        self.size = size
        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._worker_rng = WorkerRng(self.worker_config)
        self._active_buffer = SavableSampleBuffer(self.dataset, worker_config=self.worker_config)

    def __len__(self) -> int:
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample]:
        self._active_buffer.worker_start()
        it = iter(self._active_buffer.append_iter())
        while True:
            if len(self._active_buffer) >= self.size:
                pop_idx = self._worker_rng.randbelow(len(self._active_buffer))
                yield self._active_buffer.pop(pop_idx)
            else:
                try:
                    next(it)
                except StopIteration:
                    break
        while len(self._active_buffer) > 0:
            pop_idx = self._worker_rng.randbelow(len(self._active_buffer))
            yield self._active_buffer.pop(pop_idx)

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T_sample:
        return self._active_buffer.restore_sample(restore_key)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "size": self.size,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"ShuffleBufferDataset(size={self.size}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import json
import multiprocessing
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, TextIO, TypeVar

import torch.distributed
import torch.utils.data

from megatron.energon.cache import CachePool

__all__ = ("WorkerConfig",)

T = TypeVar("T")


@dataclass(slots=True, kw_only=True, eq=False)
class WorkerConfig:
    """
    Provides information about the current worker and the global configuration. This gives each
    data parallel rank its proper config. Every `rank` (up to `world_size-1`) must be used.
    If set wrong, the datasets might yield the same data or data might be missing, as data
    is split over the data parallel ranks with this config!
    You may set the same rank, if you need multiple ranks to retrieve the same data.
    """

    #: The data parallel rank/id of the current process.
    rank: int
    #: The total number of data parallel processes.
    world_size: int
    #: The number of workers per rank. May be 0 to disable worker processes.
    num_workers: int

    #: If not using all ranks for data parallel, set this to the corresponding group.
    data_parallel_group: Optional[torch.distributed.ProcessGroup] = None

    #: The id offset of the current worker. e.g. the worker may live as `worker_info.id=0`, but
    # actually yield samples for id=1 (i.e. worker_id_offset=1). Required to support restoring the
    # worker state if last emitted sample was not for worker_id=0. Required by SavableDataLoader to
    # restore the worker state. Is only set to nonzero within a worker process.
    worker_id_offset: ClassVar[int] = 0

    #: The following seed_offset is used used at two points in the code.
    # 1. The seed_offset in the worker_config that is passed to the dataset initialization, is used
    #    to set the seed for the dataset shuffling and shuffled blending (All code that uses WorkerRng).
    # 2. The worker_config passed to the data loader initialization, is used to set the seed for the
    #    torch, numpy and random libraries. This does not affect the dataset shuffling, but only the
    #    user code (e.g. code in TaskEncoder).
    seed_offset: int = 0

    #: The path to the debug file for the current worker. Should contain "{worker_id}" and "{pid}"
    # to separate the workers.
    worker_debug_path: Optional[str] = None
    #: Log level for worker logging.
    worker_log_level: int = 0
    #: The opened file for the current worker. Should not be set from outside.
    _worker_debug_file: Optional[TextIO] = None
    #: worker_id of the opened worker debug file
    _worker_debug_file_worker_id: Optional[int] = None

    #: The current sample index within the current iterating worker
    _sample_index_stack: ClassVar[Optional[List[int]]] = None
    #: The current worker config within the current iterating worker
    active_worker_config: ClassVar[Optional["WorkerConfig"]] = None

    #: The global rank override for the worker. Required for restoring samples.
    _worker_override_global_rank: ClassVar[Optional[List[int]]] = None

    #: The current cache pool for the worker.
    _cache_pool: "ClassVar[Optional[CachePool]]" = None

    def worker_activate(
        self,
        sample_index: int,
        override_global_rank: Optional[int] = None,
        cache_pool: "Optional[CachePool]" = None,
    ):
        """Activates the worker config for the current worker and sets it as actively iterating.
        Must be called before next() call on the datasets."""
        assert WorkerConfig.active_worker_config is None
        WorkerConfig._sample_index_stack = [sample_index]
        WorkerConfig.active_worker_config = self
        WorkerConfig._worker_override_global_rank = override_global_rank
        WorkerConfig._cache_pool = cache_pool

    def worker_push_sample_index(self, sample_index: int):
        """Pushes a new sample index to the sample index stack. Should be set by wrapping datasets
        before calling inners."""
        assert WorkerConfig.active_worker_config is not None
        WorkerConfig._sample_index_stack.append(sample_index)

    def worker_pop_sample_index(self):
        """Pushes a new sample index to the sample index stack. Should be set by wrapping datasets
        before calling inners."""
        assert WorkerConfig.active_worker_config is not None
        return WorkerConfig._sample_index_stack.pop()

    def worker_deactivate(self):
        """Deactivates the worker config for the current worker and deactivates it for iterating.
        Must be called after next() call on the datasets."""
        if WorkerConfig.active_worker_config is not None:
            assert len(WorkerConfig._sample_index_stack) == 1, (
                f"Sample index stack not empty: {WorkerConfig._sample_index_stack}"
            )
            WorkerConfig._sample_index_stack = None
            WorkerConfig.active_worker_config = None
            WorkerConfig._worker_override_global_rank = None

    @property
    def active_worker_sample_index(self) -> int:
        """Returns the current sample index for the actively iterating worker."""
        # Internal sample index is for the local worker. If using multiple workers per rank, this
        # must be multiplied by the number of workers and offset by the local worker index.
        return (
            WorkerConfig._sample_index_stack[-1] * max(self.num_workers, 1) + self.rank_worker_id()
        )

    @property
    def active_worker_batch_index(self) -> int:
        """Returns the current batch index for the actively iterating worker."""
        # Internal batch index is for the local worker. If using multiple workers per rank, this
        # must be multiplied by the number of workers and offset by the local worker index.
        return (
            WorkerConfig._sample_index_stack[0] * max(self.num_workers, 1) + self.rank_worker_id()
        )

    def global_rank(self) -> int:
        """Returns the global rank of this worker config but as a global rank, not
        as a rank within the data parallel group."""

        if self.data_parallel_group is None:
            return self.rank

        return torch.distributed.get_global_rank(self.data_parallel_group, self.rank)

    def __eq__(self, other):
        """Do not compare everything to check for equal config"""
        if not isinstance(other, WorkerConfig):
            return NotImplementedError()
        return all(
            [
                self.rank == other.rank,
                self.world_size == other.world_size,
                self.num_workers == other.num_workers,
            ]
        )

    @staticmethod
    def default_worker_config(
        num_workers: int = 4, data_parallel_group: Optional[torch.distributed.ProcessGroup] = None
    ) -> "WorkerConfig":
        """Returns the default worker config using torch distributed if available.
        If torch distributed is not available, a single local rank is assumed."""

        if torch.distributed.is_available() and torch.distributed.is_initialized():
            rank = torch.distributed.get_rank(data_parallel_group)
            world_size = torch.distributed.get_world_size(data_parallel_group)
        else:
            rank = 0
            world_size = 1
        return WorkerConfig(
            rank=rank,
            world_size=world_size,
            num_workers=num_workers,
            data_parallel_group=data_parallel_group,
        )

    def rank_worker_id(self) -> int:
        """Returns the self worker id within the current rank."""
        if self._worker_override_global_rank:
            assert self.worker_id_offset == 0
            return self._worker_override_global_rank % self.num_workers
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            return self.worker_id_offset
        assert worker_info.num_workers == self.num_workers
        # Apply the worker_id_offset as a left rotation of the logical worker ids.
        # This ensures that after restoring a checkpoint the first physical
        # worker (id=0) corresponds to the logical worker that should emit the
        # next sample. For example, if `worker_id_offset` is 1, logical worker
        # 1 becomes the first to emit a sample, shifting the ordering forward.
        return (worker_info.id + self.worker_id_offset) % worker_info.num_workers

    def assert_worker(self):
        """Checks if the current process is a worker (if configured so), and that the workers are
        properly configured."""
        if self.num_workers <= 1:
            assert self.rank_worker_id() == 0
        else:
            worker_info = torch.utils.data.get_worker_info()
            assert worker_info is not None, "Cannot iterate out of worker context"
            assert worker_info.num_workers == self.num_workers, (
                f"Actual number of workers for this rank ({worker_info.num_workers}) does not "
                f"match the configured number of workers ({self.num_workers})"
            )

    def global_worker_id(self, override_local_worker_id: Optional[int] = None) -> int:
        """Returns the global worker index by multiplying the rank with the number of workers.
        Alternatively, you can override the local worker id.

        Args:
            override_local_worker_id (int, optional): The local worker id to override. None means
                the current worker, which is the default.
        """
        if self._worker_override_global_rank is not None:
            assert override_local_worker_id is None
            return self._worker_override_global_rank

        if override_local_worker_id is not None:
            return self.rank * self.num_workers + override_local_worker_id
        else:
            self.assert_worker()
            return self.rank * self.num_workers + self.rank_worker_id()

    def worker_seed(self, override_local_worker_id: Optional[int] = None) -> int:
        """Returns the seed for the current worker (or a specified worker).
        Base on the current worker id and the seed offset, compute a seed.
        Alternatively, you can override the local worker id with a fixed one to
        pregenerate seeds for multiple workers.

        Args:
            override_local_worker_id (int, optional): The local worker id to override. None means
                the current worker, which is the default.
        """

        if self.num_workers == 0:
            # If we are not using workers, different ranks should still get a different seed
            global_worker_id = self.rank
        else:
            global_worker_id = self.global_worker_id(override_local_worker_id)

        seed_offset = self.seed_offset

        seed_hash = hashlib.sha1(f"{global_worker_id},{seed_offset}".encode("utf-8")).digest()

        return int.from_bytes(seed_hash, byteorder="big", signed=False) & 0xFFFFFFFF

    def config(self) -> Dict[str, Any]:
        return {
            "rank": self.rank,
            "world_size": self.world_size,
            "num_workers": self.num_workers,
            "data_parallel_group": (
                self.data_parallel_group.size() if self.data_parallel_group else None
            ),
        }

    def should_log(self, level: int) -> bool:
        return level <= self.worker_log_level

    def worker_log(self, data: dict) -> None:
        """Logs the given data to the worker debug file."""
        if self.worker_debug_path is None:
            print(json.dumps(data) + "\n", end="", flush=True)
        else:
            in_worker = torch.utils.data.get_worker_info() is not None
            # Additional "worker" with rank_worker_id=0 is the main process. All workers have +1
            # as their worker_id.
            worker_id = (
                self.rank * (self.num_workers + 1) + self.rank_worker_id() + (1 if in_worker else 0)
            )
            if self._worker_debug_file is None or self._worker_debug_file_worker_id != worker_id:
                if self._worker_debug_file is not None:
                    self._worker_debug_file.close()
                path = Path(
                    self.worker_debug_path.format(
                        worker_id=worker_id, pid=multiprocessing.current_process().ident
                    )
                )
                path.parent.mkdir(exist_ok=True, parents=True)
                self._worker_debug_file = path.open("w")
                self._worker_debug_file_worker_id = worker_id
            self._worker_debug_file.write(json.dumps(data) + "\n")
            self._worker_debug_file.flush()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import struct
import tarfile
from types import TracebackType
from typing import BinaryIO, Dict, Generator, Optional, Tuple, Type, Union

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.retry_stream import RetryReadStream

ITAR_SUFFIX = ".tar.idx"


@edataclass
class ITarSamplePointer:
    """
    Points to a sample inside some tar file on disk.
    The tar_file_id refers to the tar_filenames in the reader.
    """

    # The index of the tar file, to be matched with the tar_filenames in the reader.
    tar_file_id: int
    # The byte offset of the sample in the tar file.
    byte_offset: int
    # The size of the sample in the tar file.
    byte_size: int


class TarIndexReader:
    def __init__(self, tar_path: Union[EPath, str]):
        tar_path = EPath(tar_path)
        self.itar = (tar_path.with_suffix(ITAR_SUFFIX)).open("rb")
        self._length = len(self)

    def __getitem__(self, index: int) -> int:
        if index >= self._length or index < 0:
            raise IndexError(f"Index {index} out of range")

        if self.itar.tell() != 8 * index:
            self.itar.seek(8 * index)

        return struct.unpack("Q", self.itar.read(8))[0]

    def __iter__(self) -> Generator[int, None, None]:
        self.itar.seek(0)
        while True:
            raw = self.itar.read(8)
            if len(raw) == 0:
                break
            assert len(raw) == 8
            yield struct.unpack("Q", raw)[0]

    def __len__(self) -> int:
        self.itar.seek(0, 2)
        return self.itar.tell() // 8

    def close(self):
        self.itar.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class TarIndexWriter:
    def __init__(self, tar_path: EPath):
        self.final_name = tar_path.with_suffix(ITAR_SUFFIX)
        self.tmp_name = tar_path.with_suffix(ITAR_SUFFIX + ".tmp")
        self.itar = self.tmp_name.open("wb")

    def append(self, offset: int):
        self.itar.write(struct.pack("Q", offset))

    def close(self, finalize: bool = True):
        self.itar.close()
        if finalize:
            self.tmp_name.move(self.final_name)
        else:
            self.tmp_name.unlink()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close(finalize=exc_val is None)


class SubFileReader(BinaryIO):
    """A file-like object that reads a subfile (i.e. offset, size defined portion) of a larger
    file."""

    def __init__(self, stream: BinaryIO, offset: int, size: int):
        self.offset = offset
        self._pos = 0
        self.size = size
        self.stream = stream
        self.stream.seek(self.offset)

    def read(self, n: int = -1) -> bytes:
        if n == -1:
            n = self.size - self._pos
        else:
            n = min(n, self.size - self._pos)
        if n == 0:
            return b""
        read = self.stream.read(n)
        self._pos += len(read)
        return read

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            self._pos = offset
        elif whence == 1:
            self._pos += offset
        elif whence == 2:
            self._pos = self.size + offset
        else:
            raise ValueError("Invalid whence value")
        self._pos = max(0, min(self._pos, self.size))
        self.stream.seek(self.offset + self._pos)
        return self._pos

    def tell(self) -> int:
        return self._pos

    def __enter__(self) -> BinaryIO:
        return self

    def __exit__(
        self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType
    ) -> None:
        self.close()

    def close(self) -> None:
        self.stream.close()

    def isatty(self) -> bool:
        return False

    def seekable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False


def get_itar_byte_offset(
    path: Union[str, EPath],
    sample_offset: int = 0,
) -> int:
    """Gets the byte offset from sample offsets."""
    if sample_offset == 0:
        return 0
    with TarIndexReader(path) as itar:
        return itar[sample_offset]


@edataclass
class CacheEntry:
    tar_index_reader: TarIndexReader
    lookahead_offset: Optional[int] = None
    lookahead_byteoffset: Optional[int] = None


class CachedItarOffsetReader:
    """
    This class is a high-level wrapper around TarIndexReader that caches some
    of the recent lookups for faster access. It is designed for the case when
    you need to read multiple offsets from the same tar file or from multiple
    tar files.

    Args:
        cache_size: The number of entries to keep in the cache. By default, we keep 32.
    """

    def __init__(self, cache_size: int = 32):
        # Maps (tar_file, current_offset) -> CacheEntry
        self.tar_index_reader_cache: Dict[Tuple[str, int], CacheEntry] = {}
        self.cache_size = cache_size

    def _find_or_create_entry(
        self,
        tar_file: Union[str, "EPath"],
        sample_offset: int,
    ) -> Tuple[Tuple[str, int], CacheEntry]:
        """
        1. If we already have a key == (tar_file, sample_offset), return it.
        2. Otherwise, create a new entry (and evict if necessary).
        """
        tar_file = str(tar_file)
        key = (tar_file, sample_offset)

        # Direct hit in the cache?
        if key in self.tar_index_reader_cache:
            return key, self.tar_index_reader_cache[key]

        # We didn't find an existing entry. Create a new one.
        # Evict if needed.
        if len(self.tar_index_reader_cache) >= self.cache_size:
            self._evict_one_entry()

        new_reader = TarIndexReader(tar_file)
        cache_entry = CacheEntry(tar_index_reader=new_reader)
        self.tar_index_reader_cache[key] = cache_entry
        return key, cache_entry

    def _evict_one_entry(self):
        """
        Evict the 'oldest' item in the cache. Here we just pop the first item
        returned by iter(...) in Python 3.7+ which *should* be insertion order,
        but not strictly an LRU. For true LRU, you can use OrderedDict or similar.
        """
        oldest_key = next(iter(self.tar_index_reader_cache))
        oldest_entry = self.tar_index_reader_cache.pop(oldest_key)
        oldest_entry.tar_index_reader.close()

    def _get_itar_byte_offset_with_entry(
        self,
        cache_entry: CacheEntry,
        sample_offset: int,
    ) -> Tuple[int, int]:
        """
        Return (start_byte_offset, length_to_next),
        possibly using per-entry lookahead for speed.
        """
        tar_index_reader = cache_entry.tar_index_reader

        # If offset=0, define the result as byte offset=0 for convenience
        if sample_offset == 0:
            result_byte_offset = 0
        elif sample_offset == cache_entry.lookahead_offset:
            # Reuse the previously cached byte offset from the lookahead
            assert cache_entry.lookahead_byteoffset is not None, (
                "Lookahead offset matched but no lookahead byte offset found."
            )
            result_byte_offset = cache_entry.lookahead_byteoffset
        else:
            # Normal random access
            result_byte_offset = tar_index_reader[sample_offset]

        # Prepare the lookahead for (sample_offset+1)
        next_offset = sample_offset + 1
        try:
            cache_entry.lookahead_byteoffset = tar_index_reader[next_offset]
            cache_entry.lookahead_offset = next_offset
        except IndexError:
            cache_entry.lookahead_offset = None
            cache_entry.lookahead_byteoffset = None

        # length = difference to the next offset, or 0 if none
        if cache_entry.lookahead_byteoffset is not None:
            length = cache_entry.lookahead_byteoffset - result_byte_offset
        else:
            length = 0

        return result_byte_offset, length

    def get_itar_byte_offset(
        self,
        tar_file: Union[str, "EPath"],
        sample_offset: int = 0,
    ) -> Tuple[int, int]:
        """
        High-level API to get the byte offset and length for the given file & sample_offset.
        """

        # Find or create the suitable CacheEntry
        key, entry = self._find_or_create_entry(tar_file, sample_offset)

        # Use (and update) the per-entry lookahead logic
        result_byte_offset, length = self._get_itar_byte_offset_with_entry(entry, sample_offset)

        # Update cache entry with the new offset
        self.tar_index_reader_cache.pop(key)
        if entry.lookahead_offset is not None:
            new_key = (str(tar_file), entry.lookahead_offset)
            if new_key not in self.tar_index_reader_cache:
                self.tar_index_reader_cache[new_key] = entry
            else:
                # Already have this entry in the cache, so we can close the reader and use the existing one
                # TODO: We may actually may want to keep multiple readers open, because they may be multiple
                # sequences to the same sequence.
                entry.tar_index_reader.close()
        else:
            # No lookahead, so we can close the reader
            entry.tar_index_reader.close()

        return result_byte_offset, length


class ITarFile(tarfile.TarFile):
    """This class is a subclass of tarfile.TarFile that allows for reading a tarfile,
    with random access while keeping the file open.

    Usage:
        with open(filename, "rb") as fileobj:
            with ITarFile.open(fileobj=fileobj, mode="r:") as f:
                f.offset = 101888
                tarinfo = f.next()
                print(tarinfo.name)
                member_bytes = f.extractfile(tarinfo)

                # Read more offsets here ...
    """

    def __init__(self, *args, **kwargs):
        self.in_init = True
        try:
            super().__init__(*args, **kwargs)
        finally:
            self.in_init = False

    def next(self):
        if self.in_init:
            # Don't automatically read the first member
            return None

        if self.offset != self.fileobj.tell():
            # This prevents tarfile from reading the one byte before
            self.fileobj.seek(self.offset)

        return super().next()


@contextlib.contextmanager
def open_itar(path: Union[str, EPath], byte_offset: int = 0, byte_size: Optional[int] = None):
    """
    Open an indexed tarfile with offset and size.
    Args:
        path: Path to the tarfile to open
        byte_offset: Byte offset within the file
        byte_size: Size of the file to read

    Returns:
        The opened tarfile
    """
    path = EPath(path)

    # TODO: if tar file startswith(b"\x1f\x8b\x08") -> Seekable gzip file
    with path.open("rb") as f:
        if f.read(3) == b"\x1f\x8b\x08":
            # Open as seekable tgz
            raise ValueError("Seekable tgz not supported yet")

    if byte_offset != 0 or byte_size is not None:
        if byte_size is None:
            byte_size = path.size() - byte_offset
        with RetryReadStream(path) as stream:
            with SubFileReader(
                stream,
                offset=byte_offset,
                size=byte_size,
            ) as fileobj:
                with ITarFile.open(fileobj=fileobj, mode="r:") as f:
                    yield f
    else:
        with RetryReadStream(path) as fileobj:
            with ITarFile.open(fileobj=fileobj, mode="r:") as f:
                yield f



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import traceback
from typing import Optional

from megatron.energon.source_info import SourceInfo


def log_exception(_e: Exception, sample, sources: Optional[list[SourceInfo]] = None):
    traceback.print_exc()
    print("-" * 10)

    if sources:
        print("Sources:")
        for source in sources:
            print(
                f" - {source.dataset_path}[{source.index}] {source.shard_name}{source.file_names!r}"
            )
        print("-" * 10)

    sample_str = str(sample)
    if len(sample_str) > 400:
        sample_str = sample_str[:200] + "..." + sample_str[-200:]

    print(sample_str)

    print("-" * 10)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import inspect
import linecache
import os
import sys
import threading
import time
import traceback
from time import perf_counter
from typing import Any, Callable, Iterable, Iterator, Optional, TypeVar

import torch
from torch.distributed._shard.sharded_tensor import ShardedTensorBase

# For the watch_iter type
T = TypeVar("T")

# Maximum length of a single object string to print.
PRINT_LOCAL_MAX_LENGTH = 250


class Watchdog:
    """
    A watchdog timer that:
      - can be 'enabled' or 'disabled' by presence/absence of a deadline,
      - resets automatically when 'enable()' is called,
      - can be used as a context manager,
      - can wrap an iterator to watch only the time for 'next()' calls,
      - attempts a two-phase shutdown on callback error:
         1) sys.exit(1) for graceful,
         2) if still alive after 10s, os._exit(1).
    """

    def __init__(
        self,
        timeout: float,
        initial_timeout: Optional[float] = None,
        callback: Optional[Callable[[], None]] = None,
        dump_stacks: bool = True,
        enabled: bool = True,
    ) -> None:
        """
        Args:
            timeout: Number of seconds before the watchdog fires if not reset/disabled.
            initial_timeout: Number of seconds before the watchdog fires in the first iteration.
            callback: Optional function to call upon timeout.
            dump_stacks: If True, print full stack traces for all threads on timeout (except watchdog's own thread).
            enabled: If False, watchdog starts disabled until enable() is called.
        """
        self._timeout = timeout
        self._initial_timeout = initial_timeout
        self._callback = callback
        self._dump_stacks = dump_stacks
        self._is_first_iteration = True

        # If _deadline is None, the watchdog is disabled.
        # Otherwise, _deadline = time.time() + _timeout if enabled.
        if enabled:
            self._deadline: Optional[float] = perf_counter() + self._get_next_timeout()
        else:
            self._deadline = None

        self._stop = False  # signals permanent shutdown (finish)

        # Condition variable to manage state changes
        self._cv = threading.Condition()
        # Background thread (daemon) that monitors timeouts
        self._worker_thread = threading.Thread(target=self._worker, daemon=True)
        self._worker_thread.start()

    def _get_next_timeout(self) -> float:
        if self._is_first_iteration:
            self._is_first_iteration = False
            return self._initial_timeout if self._initial_timeout is not None else self._timeout
        else:
            return self._timeout

    def _worker(self) -> None:
        """
        Background thread that periodically checks if the watchdog has expired.
        Once it times out or is told to stop, it exits.
        """
        while True:
            with self._cv:
                if self._stop:
                    # finish() was called; end the worker.
                    return

                if self._deadline is None:
                    # Disabled; no deadline. Just wait a bit, then re-check.
                    self._cv.wait(timeout=1.0)
                    continue

                remaining = self._deadline - perf_counter()
                if remaining <= 0:
                    # We have timed out
                    self._on_timeout()
                    return
                else:
                    # Wait until either the deadline or a state change
                    self._cv.wait(timeout=remaining)

    def _on_timeout(self) -> None:
        """
        Called exactly once if the watchdog times out.
        1) Optionally dumps stacks,
        2) Calls user callback,
        3) If callback raises an error,
           - print traceback,
           - sys.exit(1),
           - fallback to os._exit(1) after 10s if process not terminated.
        """
        watchdog_thread_id = threading.get_ident()

        # 1) Dump stacks if requested
        if self._dump_stacks:
            print("Watchdog triggered: Dumping thread stacks")
            self._print_all_thread_stacks(skip_thread_id=watchdog_thread_id)

        # 2) Call user callback
        if self._callback:
            try:
                self._callback()
            except Exception:
                # Print the traceback
                traceback.print_exc()

                # Start a background kill-switch after 10 seconds
                def force_exit_after_delay() -> None:
                    time.sleep(10)
                    os._exit(1)

                killer = threading.Thread(target=force_exit_after_delay, daemon=True)
                killer.start()

                # Attempt graceful shutdown
                sys.exit(1)

    def _print_all_thread_stacks(self, skip_thread_id: Optional[int] = None) -> None:
        """
        Dump stacks of all threads in a style reminiscent of py-spy, from
        innermost (current) to outermost. Skip the watchdog's own thread if given.

        Args:
            skip_thread_id: If given, skip this thread's stack.
        """

        frames = sys._current_frames()  # thread_id -> frame
        # We gather known threads to print their names
        all_threads = {t.ident: t for t in threading.enumerate()}

        for thread_id, frame in frames.items():
            if skip_thread_id is not None and thread_id == skip_thread_id:
                continue

            thread = all_threads.get(thread_id)
            thread_name = thread.name if thread else f"Unknown-{thread_id}"
            print(f'Thread {thread_id}: "{thread_name}"')

            # Build the stack from current (innermost) to outermost
            stack_frames = []
            f = frame
            while f is not None:
                stack_frames.append(f)
                f = f.f_back

            for fr in stack_frames:
                code = fr.f_code
                func_name = code.co_name
                filename = code.co_filename
                lineno = fr.f_lineno

                print(f"    {func_name} ({filename}:{lineno})")

                # Attempt to read the actual line of source
                line = linecache.getline(filename, lineno).rstrip()
                if line:
                    print(f"        > {line}")

                # Show arguments and locals
                arg_info = inspect.getargvalues(fr)
                arg_names = arg_info.args
                varargs = arg_info.varargs
                varkw = arg_info.keywords
                local_vars = arg_info.locals

                # Separate out the arguments
                arg_dict = {}
                for arg in arg_names:
                    if arg in local_vars:
                        arg_dict[arg] = local_vars[arg]
                if varargs and varargs in local_vars:
                    arg_dict["*" + varargs] = local_vars[varargs]
                if varkw and varkw in local_vars:
                    arg_dict["**" + varkw] = local_vars[varkw]

                if arg_dict:
                    print("        Arguments:")
                    for k, v in arg_dict.items():
                        print(f"            {k}: {repr_short(v)}")

                other_locals = {k: v for k, v in local_vars.items() if k not in arg_dict}
                if other_locals:
                    print("        Locals:")
                    for k, v in other_locals.items():
                        print(f"            {k}: {repr_short(v)}")

            print(flush=True)

    def reset(self) -> None:
        """
        Reset the watchdog timer (push out deadline by `timeout` seconds),
        but only if currently enabled (i.e., _deadline is not None).
        """
        with self._cv:
            if self._deadline is not None:
                self._deadline = perf_counter() + self._timeout
                self._cv.notify()

    def enable(self) -> None:
        """
        Enable (or re-enable) the watchdog. Always resets the deadline to
        `time.time() + timeout`.
        """
        with self._cv:
            self._deadline = perf_counter() + self._get_next_timeout()
            self._cv.notify()

    def disable(self) -> None:
        """
        Disable the watchdog (no timeout will fire until re-enabled).
        """
        with self._cv:
            self._deadline = None
            self._cv.notify()

    def finish(self) -> None:
        """
        Permanently stop the watchdog thread and disarm the timer.
        After calling finish(), you cannot re-enable this watchdog.
        """
        with self._cv:
            self._stop = True
            self._cv.notify()
        self._worker_thread.join()

    def __enter__(self) -> "Watchdog":
        # If currently disabled, calling enable() will also reset the timer
        if self._deadline is None:
            self.enable()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # End the watchdog on context exit
        self.finish()

    def watch_iter(self, iterable: Iterable[T]) -> Iterator[T]:
        """
        Wrap an iterable so that each 'next()' call is watched by the watchdog,
        but the time in between iterations is not watched. Usage:

            wd = Watchdog(timeout=3, enabled=False)
            for item in wd.watch_iter(generator()):
                # processing item not timed by the watchdog
                pass

        This pattern:
          - enable() -> sets/extends deadline
          - next(...) -> measured portion
          - disable() -> stops timer

        Args:
            iterable: The iterable to wrap and watch.

        Returns:
            An iterator that wraps the input iterable and watches for timeouts.
        """

        it = iter(iterable)
        while True:
            # Automatically resets timer
            self.enable()
            try:
                item = next(it)
            except StopIteration:
                self.disable()
                break
            except:
                # On any error, disable watchdog before re-raising
                self.disable()
                raise
            else:
                self.disable()
                yield item


def repr_short(obj: Any) -> str:
    """
    Return a short repr of an object.
    """
    if isinstance(obj, torch.Tensor):
        if isinstance(obj, ShardedTensorBase) or obj.is_cuda:
            return "<CUDA tensor>"

    s = repr(obj)
    if len(s) > PRINT_LOCAL_MAX_LENGTH:
        s = s[: PRINT_LOCAL_MAX_LENGTH // 2] + "..." + s[-PRINT_LOCAL_MAX_LENGTH // 2 :]
    return s


if __name__ == "__main__":
    # Example usage

    def my_callback() -> None:
        print("Watchdog timed out in callback.")
        # Demonstrate an error
        raise ValueError("Example error from callback.")

    print("Simple usage example:")
    wd = Watchdog(timeout=2, callback=my_callback, enabled=True)
    print("Sleeping 3s so the watchdog times out.")
    time.sleep(30)
    # Because we never reset or finish, the watchdog should fire and
    # forcibly exit, after printing the traceback and stack dumps.
    print("You won't see this line if the watchdog fired first.")



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import List, Optional, Union

import torch

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib.epath import EPath
from megatron.energon.errors import warn_deprecated
from megatron.energon.flavors.base_dataset import Sample
from megatron.energon.flavors.webdataset import DefaultDecoderWebdatasetFactory


@edataclass
class OCRSample(Sample):
    """Sample type for optical character recognition."""

    #: The input image tensor in the shape (C, H, W)
    image: torch.Tensor
    #: The text contained in the image
    text: str
    #: The bounding boxes of the blocks in the image float(N, 4|5<x, y, w, h>)
    block_boxes: Optional[torch.Tensor] = None
    #: The classes of the blocks in the image int(N, 1<block_class>)
    block_classes: Optional[Union[torch.Tensor, List[str]]] = None
    #: The text contained in each block (N,)
    block_text: Optional[List[str]] = None
    #: The bounding boxes of the lines in the image float(N, 4|5<x, y, w, h[, confidence]>)
    lines_boxes: Optional[torch.Tensor] = None
    #: The text contained in each line (N,)
    lines_text: Optional[List[str]] = None
    #: The bounding boxes of the words in the image float(N, 4|5<x, y, w, h[, confidence]>)
    words_boxes: Optional[torch.Tensor] = None
    #: The text contained in each word (N,)
    words_text: Optional[List[str]] = None
    #: The bounding boxes of the chars in the image float(N, 4|5<x, y, w, h[, confidence]>)
    chars_boxes: Optional[torch.Tensor] = None
    #: The character contained in each char (N,)
    chars_text: Optional[List[str]] = None


class OCRWebdataset(DefaultDecoderWebdatasetFactory[OCRSample]):
    __sample_type__ = OCRSample

    def __init__(self, path: EPath, **kwargs):
        warn_deprecated(
            f"{type(self)} is deprecated, use the default instead and set the sample_type:\n"
            f"To convert, update your {path}/.nv-meta/dataset.yaml to:\n"
            f"# remove top-level __module__ and __class__\n"
            f"sample_type:\n"
            f"  __module__: megatron.energon\n"
            f"  __class__: {self.__sample_type__.__name__}\n"
            f"# Keep the remaining content"
        )
        super().__init__(path, **kwargs)
